from collections import defaultdict
from functools import partial
from typing import Any, Literal, TypeAlias
import weakref

import numpy as np
import pylinalg as la
from rendercanvas.base import log_exception

try:
    from imgui_bundle import imgui
except ImportError:
    IMGUI = False
else:
    IMGUI = True

import pygfx

from .features import (
    BufferManager,
    Deleted,
    Name,
    Offset,
    Rotation,
    Alpha,
    AlphaMode,
    Visible,
)
from ._axes import Axes

HexStr: TypeAlias = str

# dict that holds all world objects for a given python kernel/session
# Graphic objects only use proxies to WorldObjects
WORLD_OBJECTS: dict[HexStr, pygfx.WorldObject] = dict()  #: {hex id str: WorldObject}


PYGFX_EVENTS = [
    "key_down",
    "key_up",
    "pointer_down",
    "pointer_move",
    "pointer_up",
    "pointer_enter",
    "pointer_leave",
    "click",
    "double_click",
    "wheel",
    "close",
    "resize",
]


class Graphic:
    _features: dict[str, type] = dict()

    def __init_subclass__(cls, **kwargs):

        # set of all features
        cls._features = {
            **cls._features,
            "name": Name,
            "offset": Offset,
            "rotation": Rotation,
            "alpha": Alpha,
            "alpha_mode": AlphaMode,
            "visible": Visible,
            "deleted": Deleted,
        }
        super().__init_subclass__(**kwargs)

    def __init__(
        self,
        name: str = None,
        offset: np.ndarray | list | tuple = (0.0, 0.0, 0.0),
        rotation: np.ndarray | list | tuple = (0.0, 0.0, 0.0, 1.0),
        alpha: float = 1.0,
        alpha_mode: str = "auto",
        visible: bool = True,
        metadata: Any = None,
    ):
        """

        Parameters
        ----------
        name: str, optional
            name this graphic to use it as a key to access from the plot

        offset: (float, float, float), default (0., 0., 0.)
            (x, y, z) vector to offset this graphic from the origin

        rotation: (float, float, float, float), default (0, 0, 0, 1)
            rotation quaternion

        alpha: (float), default 1.0
            The global alpha value, i.e. opacity, of the graphic.

        The alpha value for the colors. If you make your a graphic transparent, consider setting ``alpha_mode``
            to 'blend' or 'weighted_blend' so it won't write to the depth buffer.

        alpha_mode: (str), default "auto",
            The alpha-mode, e.g. 'auto', 'blend', 'weighted_blend', 'solid', or 'dither'.

            Modes for method “opaque” (overwrites the value in the output texture):

                * “solid”: alpha is ignored.
                * “solid_premul”: the alpha is multipled with the color (making it darker).

            Modes for method “blended” (per-fragment blending, a.k.a. compositing):

                * “auto”: classic alpha blending, with depth_write defaulting to True. See note below.
                * “blend”: classic alpha blending using the over-operator. depth_write defaults to False.
                * “add”: additive blending that adds the fragment color, multiplied by alpha.
                * “subtract”: subtractuve blending that removes the fragment color.
                * “multiply”: multiplicative blending that multiplies the fragment color.

            Modes for method “weighted” (order independent blending):

                * “weighted_blend”: weighted blended order independent transparency.
                * “weighted_solid”: fragments are combined based on alpha, but the final alpha is always 1. Great for e.g. image stitching.

            Modes for method “stochastic” (alpha represents the chance of a fragment being visible):

                * “dither”: stochastic transparency with blue noise. This mode handles order-independent transparency exceptionally well, but it produces results that can look somewhat noisy.
                * “bayer”: stochastic transparency with an 8x8 Bayer pattern.


            For details see https://docs.pygfx.org/stable/transparency.html

        visible: (bool), default True
            Whether the graphic is visible.

        metadata: Any, optional
            metadata attached to this Graphic, this is for the user to manage

        """
        if (name is not None) and (not isinstance(name, str)):
            raise TypeError("Graphic `name` must be of type <str>")

        self.metadata = metadata
        self.registered_callbacks = dict()

        # store hex id str of Graphic instance mem location
        self._fpl_address: HexStr = hex(id(self))

        self._plot_area = None

        # event handlers
        self._event_handlers = defaultdict(set)

        # maps callbacks to their partials
        self._event_handler_wrappers = defaultdict(set)

        # all the common features
        self._name = Name(name)
        self._deleted = Deleted(False)
        self._rotation = Rotation(rotation)
        self._offset = Offset(offset)
        self._alpha = Alpha(alpha)
        self._alpha_mode = AlphaMode(alpha_mode)
        self._visible = Visible(visible)
        self._block_events = False

        self._axes: Axes = None

        self._right_click_menu = None

    @property
    def supported_events(self) -> tuple[str]:
        """events supported by this graphic"""
        return (*tuple(self._features.keys()), *PYGFX_EVENTS)

    @property
    def name(self) -> str | None:
        """Graphic name"""
        return self._name.value

    @name.setter
    def name(self, value: str):
        self._name.set_value(self, value)

    @property
    def offset(self) -> np.ndarray:
        """Offset position of the graphic, array: [x, y, z]"""
        return self._offset.value

    @offset.setter
    def offset(self, value: np.ndarray | list | tuple):
        self._offset.set_value(self, value)

    @property
    def rotation(self) -> np.ndarray:
        """Orientation of the graphic as a quaternion"""
        return self._rotation.value

    @rotation.setter
    def rotation(self, value: np.ndarray | list | tuple):
        self._rotation.set_value(self, value)

    @property
    def alpha(self) -> float:
        """The opacity of the graphic"""
        return self._alpha.value

    @alpha.setter
    def alpha(self, value: float):
        self._alpha.set_value(self, value)

    @property
    def alpha_mode(self) -> str:
        """How the alpha is handled by the renderer"""
        return self._alpha_mode.value

    @alpha_mode.setter
    def alpha_mode(self, value: str):
        self._alpha_mode.set_value(self, value)

    @property
    def visible(self) -> bool:
        """Whether the graphic is visible"""
        return self._visible.value

    @visible.setter
    def visible(self, value: bool):
        self._visible.set_value(self, value)

    @property
    def deleted(self) -> bool:
        """used to emit an event after the graphic is deleted"""
        return self._deleted.value

    @deleted.setter
    def deleted(self, value: bool):
        self._deleted.set_value(self, value)

    @property
    def block_events(self) -> bool:
        """Used to block events for a graphic and prevent recursion."""
        return self._block_events

    @block_events.setter
    def block_events(self, value: bool):
        self._block_events = value

    @property
    def world_object(self) -> pygfx.WorldObject:
        """Associated pygfx WorldObject. Always returns a proxy, real object cannot be accessed directly."""
        # We use weakref to simplify garbage collection
        return weakref.proxy(WORLD_OBJECTS[hex(id(self))])

    def _set_world_object(self, wo: pygfx.WorldObject):
        WORLD_OBJECTS[self._fpl_address] = wo

        wo.visible = self.visible
        if "Image" in self.__class__.__name__:
            # Image and ImageVolume use tiling and share one material
            self._material.opacity = self.alpha
            self._material.alpha_mode = self.alpha_mode

        if wo.material is not None:
            wo.material.opacity = self.alpha
            wo.material.alpha_mode = self.alpha_mode

        # set offset if it's not (0., 0., 0.)
        if not all(wo.world.position == self.offset):
            self.offset = self.offset

        # set rotation if it's not (0., 0., 0., 1.)
        if not all(wo.world.rotation == self.rotation):
            self.rotation = self.rotation

    @property
    def event_handlers(self) -> list[tuple[str, callable, ...]]:
        """
        Registered event handlers. Read-only use ``add_event_handler()``
        and ``remove_event_handler()`` to manage callbacks
        """
        return list(self._event_handlers.items())

    def add_event_handler(self, *args):
        """
        Register an event handler. Can also be used as a decorator.

        Parameters
        ----------
        callback: callable, the first argument
            Event handler, must accept a single event  argument

        *types: strings
            event types, ex: "click", "data", "colors", "pointer_down"

            ``supported_events`` will return a tuple of all event type strings that this graphic supports.
            See the user guide in the documentation for more information on events.

        Example
        -------

        .. code-block:: py

            def my_handler(event):
                print(event)

            graphic.add_event_handler(my_handler, "pointer_up", "pointer_down")

        Decorator usage example:

        .. code-block:: py

            @graphic.add_event_handler("click")
            def my_handler(event):
                print(event)
        """

        decorating = not callable(args[0])
        callback = None if decorating else args[0]
        types = args if decorating else args[1:]

        unsupported_events = [t for t in types if t not in self.supported_events]

        if len(unsupported_events) > 0:
            raise TypeError(
                f"unsupported events passed: {unsupported_events} for {self.__class__.__name__}\n"
                f"`graphic.events` will return a tuple of supported events"
            )

        def decorator(_callback):
            _callback_wrapper = partial(
                self._handle_event, _callback
            )  # adds graphic instance as attribute and other things

            for t in types:
                # add to our record
                self._event_handlers[t].add(_callback)

                if t in self._features.keys():
                    # fpl feature event
                    feature = getattr(self, f"_{t}")
                    feature.add_event_handler(_callback_wrapper)
                else:
                    # wrap pygfx event
                    self.world_object._event_handlers[t].add(_callback_wrapper)

                # keep track of the partial too
                self._event_handler_wrappers[t].add((_callback, _callback_wrapper))
            return _callback

        if decorating:
            return decorator

        return decorator(callback)

    def clear_event_handlers(self):
        """clear all event handlers added to this graphic"""
        for ev, handlers in self.event_handlers:
            handlers = list(handlers)
            for h in handlers:
                self.remove_event_handler(h, ev)

    def _handle_event(self, callback, event: pygfx.Event):
        """Wrap pygfx event to add graphic to pick_info"""
        event.graphic = self

        if self.block_events:
            return

        if event.type in self._features:
            # for feature events
            event._target = self.world_object

        with log_exception(f"Error during handling {event.type} event"):
            callback(event)

    def remove_event_handler(self, callback, *types):
        """
        remove a registered event handler

        Parameters
        ----------
        callback: callable
            event handler that has been added

        *types: strings
            event types that were registered with the given callback

        Example
        -------

        .. code-block:: py

            # define event handler
            def my_handler(event):
                print(event)

            # add event handler
            graphic.add_event_handler(my_handler, "pointer_up", "pointer_down")

            # remove event handler
            graphic.remove_event_handler(my_handler, "pointer_up", "pointer_down")

        """

        # remove from our record first
        for t in types:
            for wrapper_map in self._event_handler_wrappers[t]:
                # TODO: not sure if we can handle this mapping in a better way
                if wrapper_map[0] == callback:
                    wrapper = wrapper_map[1]
                    self._event_handler_wrappers[t].remove(wrapper_map)
                    break
            else:
                raise KeyError(
                    f"event type: {t} with callback: {callback} is not registered"
                )

            self._event_handlers[t].remove(callback)
            # remove callback wrapper from world object if pygfx event
            if t in PYGFX_EVENTS:
                self.world_object.remove_event_handler(wrapper, t)
            else:
                feature = getattr(self, f"_{t}")
                feature.remove_event_handler(wrapper)

    def _fpl_add_plot_area_hook(self, plot_area):
        self._plot_area = plot_area

    def __repr__(self):
        rval = f"{self.__class__.__name__}"
        if self.name is not None:
            return f"'{self.name}': {rval}"
        else:
            return rval

    def _fpl_prepare_del(self):
        """
        Cleans up the graphic in preparation for __del__(), such as removing event handlers from
        plot renderer, feature event handlers, etc.

        Optionally implemented in subclasses
        """
        # remove axes if added to this graphic
        if self._axes is not None:
            self._plot_area.scene.remove(self._axes)
            self._plot_area.remove_animation(self._update_axes)
            self._axes.world_object.clear()

        # signal that a deletion has been requested
        self.deleted = True

        # clear event handlers
        self.clear_event_handlers()

        # clear any attached event handlers and animation functions
        for attr in dir(self):
            try:
                method = getattr(self, attr)
            except:
                continue

            if not callable(method):
                continue

            for ev_type in PYGFX_EVENTS:
                try:
                    self._plot_area.renderer.remove_event_handler(method, ev_type)
                except (KeyError, TypeError):
                    pass

            try:
                self._plot_area.remove_animation(method)
            except KeyError:
                pass

        for child in self.world_object.children:
            child._event_handlers.clear()

        self.world_object._event_handlers.clear()

    def __del__(self):
        # remove world object if created
        # world object does not exist if an exception was raised during __init__ which is why this check exists
        WORLD_OBJECTS.pop(hex(id(self)), None)

    def rotate(self, alpha: float, axis: Literal["x", "y", "z"] = "y"):
        """Rotate the Graphic with respect to the world.

        Parameters
        ----------
        alpha :
            Rotation angle in radians.
        axis :
            Rotation axis label.
        """
        if axis == "x":
            rot = la.quat_from_euler((alpha, 0), order="XY")
        elif axis == "y":
            rot = la.quat_from_euler((0, alpha), order="XY")
        elif axis == "z":
            rot = la.quat_from_euler((0, alpha), order="XZ")
        else:
            raise ValueError(
                f"`axis` must be either `x`, `y`, or `z`. `{axis}` provided instead!"
            )
        self.rotation = la.quat_mul(rot, self.rotation)

    @property
    def axes(self) -> Axes:
        return self._axes

    def add_axes(self):
        """Add axes onto this Graphic"""
        if self._axes is not None:
            raise AttributeError("Axes already added onto this graphic")

        self._axes = Axes(self._plot_area, offset=self.offset, grids=False)
        self._axes.world_object.local.rotation = self.world_object.local.rotation

        self._plot_area.scene.add(self.axes.world_object)
        self._axes.update_using_bbox(self.world_object.get_world_bounding_box())

    @property
    def right_click_menu(self):
        return self._right_click_menu

    @right_click_menu.setter
    def right_click_menu(self, menu):
        if not IMGUI:
            raise ImportError(
                "imgui is required to set right-click menus:\npip install imgui_bundle"
            )

        self._right_click_menu = menu
        menu.owner = self

    def _fpl_request_right_click_menu(self):
        pass

    def _fpl_close_right_click_menu(self):
        pass



from collections import OrderedDict
from typing import *

import numpy as np
import cmap as cmap_lib

from pygfx import Texture, Color


cmap_catalog = cmap_lib.Catalog()

COLORMAPS = sorted(
    [
        "viridis",
        "plasma",
        "inferno",
        "magma",
        "cividis",
        "Greys",
        "Purples",
        "Blues",
        "Greens",
        "Oranges",
        "Reds",
        "tol:YlOrBr",
        "YlOrRd",
        "OrRd",
        "PuRd",
        "RdPu",
        "BuPu",
        "GnBu",
        "PuBu",
        "YlGnBu",
        "PuBuGn",
        "BuGn",
        "YlGn",
        "binary",
        "gist_yarg",
        "gist_gray",
        "gray",
        "bone",
        "pink",
        "spring",
        "summer",
        "autumn",
        "winter",
        "cool",
        "Wistia",
        "hot",
        "afmhot",
        "gist_heat",
        "matlab:copper",
        "PiYG",
        "tol:PRGn",
        "BrBG",
        "PuOr",
        "RdGy",
        "vispy:RdBu",
        "RdYlBu",
        "RdYlGn",
        "Spectral",
        "coolwarm",
        "bwr",
        "seismic",
        "berlin",
        "vanimo",
        "twilight",
        "twilight_shifted",
        "hsv",
        "Pastel1",
        "Pastel2",
        "Paired",
        "Accent",
        "Dark2",
        "Set1",
        "Set2",
        "Set3",
        "tab10",
        "tab20",
        "tab20b",
        "tab20c",
        "flag",
        "prism",
        "gnuplot:ocean",
        "gist_earth",
        "terrain",
        "gist_stern",
        "gnuplot",
        "gnuplot2",
        "CMRmap",
        "cubehelix",
        "brg",
        "gist_rainbow",
        "yorick:rainbow",
        "jet",
        "turbo",
        "nipy_spectral",
        "gist_ncar",
    ]
)

SEQUENTIAL_CMAPS = list()
QUALITATIVE_CMAPS = list()
CYCLIC_CMAPS = list()
DIVERGING_CMAPS = list()
MISC_CMAPS = list()


for name in COLORMAPS:
    _colormap = cmap_lib.Colormap(name)
    match _colormap.category:
        case "sequential":
            if _colormap.interpolation == "nearest":
                continue
            SEQUENTIAL_CMAPS.append(name)
        case "cyclic":
            if _colormap.interpolation == "nearest":
                continue
            CYCLIC_CMAPS.append(name)
        case "diverging":
            if _colormap.interpolation == "nearest":
                continue
            DIVERGING_CMAPS.append(name)
        case "qualitative":
            QUALITATIVE_CMAPS.append(name)
        case "miscellaneous":
            if _colormap.interpolation == "nearest":
                continue
            MISC_CMAPS.append(name)


COLORMAP_NAMES = {
    "sequential": sorted(SEQUENTIAL_CMAPS),
    "cyclic": sorted(CYCLIC_CMAPS),
    "diverging": sorted(DIVERGING_CMAPS),
    "qualitative": sorted(QUALITATIVE_CMAPS),
    "miscellaneous": sorted(MISC_CMAPS),
}


def get_cmap(name: str, alpha: float = 1.0, gamma: float = 1.0) -> np.ndarray:
    """
    Get a colormap as numpy array

    Parameters
    ----------
    name: str
        name of colormap
    alpha: float
        alpha, 0.0 - 1.0
    gamma: float
        gamma, 0.0 - 1.0

    Returns
    -------
    np.ndarray
        [n_colors, 4], i.e. [n_colors, RGBA]

    """

    cmap = cmap_lib.Colormap(name).lut(256, gamma=gamma)
    cmap[:, -1] = alpha
    return cmap.astype(np.float32)


def make_colors(n_colors: int, cmap: str, alpha: float = 1.0) -> np.ndarray:
    """
    Get colors from a colormap. The returned colors are uniformly spaced, except
    for qualitative colormaps where they are returned subsequently.

    Parameters
    ----------
    n_colors: int
        number of colors to get

    cmap: str
        name of colormap

    alpha: float, default 1.0
        alpha value

    Returns
    -------
    np.ndarray
        shape is [n_colors, 4], where the last dimension is RGBA

    """

    cm = cmap_lib.Colormap(cmap)

    # can also use cm.category == "qualitative", but checking for non-interpolated
    # colormaps is a bit more general.  (and not all "custom" colormaps will be
    # assigned a category)
    if cm.interpolation == "nearest":
        max_colors = len(cm.color_stops)
        if n_colors > max_colors:
            raise ValueError(
                f"You have requested <{n_colors}> colors but only <{max_colors}> exist for the "
                f"chosen cmap: <{cmap}>"
            )
        return np.asarray(cm.color_stops, dtype=np.float32)[:n_colors, 1:]

    cm_ixs = np.linspace(0, 255, n_colors, dtype=int)
    return cm(cm_ixs).astype(np.float32)


def get_cmap_texture(name: str, alpha: float = 1.0) -> Texture:
    return Texture(get_cmap(name, alpha), dim=1)


def make_colors_dict(labels: Sequence, cmap: str, **kwargs) -> OrderedDict:
    """
    Get a dict for mapping labels onto colors.

    Parameters
    ----------
    labels: Sequence[Any]
        labels for creating a colormap. Order is maintained if it is a list of unique elements.

    cmap: str
        name of colormap

    **kwargs
        passed to make_colors()

    Returns
    -------
    OrderedDict
        keys are labels, values are colors

    Examples
    --------

    .. code-block:: python

        from fastplotlib.utils import get_colors_dict

        labels = ["l1", "l2", "l3"]
        labels_cmap = get_colors_dict(labels, cmap="tab10")

        # illustration of what the `labels_cmap` dict would look like:
        # keep in mind that the tab10 cmap was chosen here

        {
            "l1": <RGBA array for the blue 'tab10' color>,
            "l2": <RGBA array for the orange 'tab10' color>,
            "l3": <RGBA array for the green 'tab10' color>,
        }

        # another example with a non-qualitative cmap
        labels_cmap_seismic = get_colors_dict(labels, cmap="bwr")

        {
            "l1": <RGBA array for the blue 'bwr' color>,
            "l2": <RGBA array for the white 'bwr' color>,
            "l3": <RGBA array for the red 'bwr' color>,
        }

    """
    if not len(set(labels)) == len(labels):
        labels = list(set(labels))
    else:
        labels = list(labels)

    colors = make_colors(len(labels), cmap, **kwargs)

    return OrderedDict(zip(labels, colors))


def quick_min_max(data: np.ndarray, max_size=1e6) -> tuple[float, float]:
    """
    Estimate the min/max values of *data* by subsampling relative to the size of each dimension in the array.

    Parameters
    ----------
    data: np.ndarray or array-like

    max_size : int, optional
        subsamples data array to this max size

    Returns
    -------
    (float, float)
        (min, max) estimate
    """

    if hasattr(data, "min") and hasattr(data, "max"):
        # if value is pre-computed
        if isinstance(data.min, (float, int, np.number)) and isinstance(
            data.max, (float, int, np.number)
        ):
            return data.min, data.max

    data = subsample_array(data, max_size=max_size)

    return float(np.nanmin(data)), float(np.nanmax(data))


def make_pygfx_colors(colors, n_colors):
    """
    Parse and make colors array using pyfx.Color

    Parameters
    ----------
    colors: str, list, tuple, or np.ndarray
        pygfx parseable color

    n_colors: int
        number of repeats of the color

    Returns
    -------
    np.ndarray
        shape is [n_colors, 4], i.e. [n_colors, RGBA]
    """

    c = Color(colors)
    colors_array = np.repeat(np.array([c]), n_colors, axis=0)

    return colors_array


def calculate_figure_shape(n_subplots: int) -> tuple[int, int]:
    """
    Returns ``(n_rows, n_cols)`` from given number of subplots ``n_subplots``
    """
    sr = np.sqrt(n_subplots)

    return (int(np.round(sr)), int(np.ceil(sr)))


def normalize_min_max(a):
    """normalize an array between 0 - 1"""
    if np.unique(a).size == 1:
        return np.zeros(a.size)

    return (a - np.min(a)) / (np.max(a - np.min(a)))


def parse_cmap_values(
    n_colors: int,
    cmap_name: str,
    transform: np.ndarray | list[int | float] = None,
) -> np.ndarray:
    """

    Parameters
    ----------
    n_colors: int
        number of graphics in collection

    cmap_name: str
        colormap name

    transform: np.ndarray | List[int | float], optional
        cmap transform
    Returns
    -------

    """
    if transform is None:
        colors = make_colors(n_colors, cmap_name)
        return colors

    else:
        if not isinstance(transform, np.ndarray):
            transform = np.array(transform)

        # use the of the cmap_transform to set the color of the corresponding data
        # each individual data[i] has its color based on the transform values
        if len(transform) != n_colors:
            raise ValueError(
                f"len(cmap_values) != len(data): {len(transform)} != {n_colors}"
            )

        colormap = get_cmap(cmap_name)

        n_colors = colormap.shape[0] - 1

        # can also use cm.category == "qualitative"
        if cmap_lib.Colormap(cmap_name).interpolation == "nearest":

            # check that cmap_values are <int> and within the number of colors `n_colors`

            # do not scale, use directly
            if not np.issubdtype(transform.dtype, np.integer):
                raise TypeError(
                    f"<int> `cmap_transform` values should be used with qualitative colormaps, "
                    f"the dtype you have passed is {transform.dtype}"
                )
            if max(transform) > n_colors:
                raise IndexError(
                    f"You have chosen the qualitative colormap <'{cmap_name}'> which only has "
                    f"<{n_colors}> colors, which is lower than the max value of your `cmap_transform`."
                    f"Choose a cmap with more colors, or a non-quantitative colormap."
                )
            norm_cmap_values = transform
        else:
            # scale between 0 - n_colors so we can just index the colormap as a LUT
            norm_cmap_values = (normalize_min_max(transform) * n_colors).astype(int)

        # use colormap as LUT to map the cmap_values to the colormap index
        colors = np.vstack([colormap[val] for val in norm_cmap_values])

        return colors


def subsample_array(
    arr: np.ndarray, max_size: int = 1e6, ignore_dims: Sequence[int] | None = None
):
    """
    Subsamples an input array while preserving its relative dimensional proportions.

    The dimensions (shape) of the array can be represented as:

    .. math::

        [d_1, d_2, \\dots d_n]

    The product of the dimensions can be represented as:

    .. math::

        \\prod_{i=1}^{n} d_i

    To find the factor ``f`` by which to divide the size of each dimension in order to
    get max_size ``s`` we must solve for ``f`` in the following expression:

    .. math::

        \\prod_{i=1}^{n} \\frac{d_i}{\\mathbf{f}} = \\mathbf{s}

    The solution for ``f`` is is simply the nth root of the product of the dims divided by the max_size
    where n is the number of dimensions

    .. math::

        \\mathbf{f} = \\sqrt[n]{\\frac{\\prod_{i=1}^{n} d_i}{\\mathbf{s}}}

    Parameters
    ----------
    arr: np.ndarray
        input array of any dimensionality to be subsampled.

    max_size: int, default 1e6
        maximum number of elements in subsampled array

    ignore_dims: Sequence[int], optional
        List of dimension indices to exclude from subsampling (i.e. retain full resolution).
        For example, `ignore_dims=[0]` will avoid subsampling along the first axis.

    Returns
    -------
    np.ndarray
        subsample of the input array
    """
    full_shape = np.array(arr.shape, dtype=np.uint64)
    if np.prod(full_shape) <= max_size:
        return arr[:]  # no need to subsample if already below the threshold

    # get factor by which to divide all dims
    f = np.power((np.prod(full_shape) / max_size), 1.0 / arr.ndim)

    # new shape for subsampled array
    ns = np.floor(np.array(full_shape) / f).clip(min=1)

    # get the step size for the slices
    slices = list(
        slice(None, None, int(s)) for s in np.floor(full_shape / ns).astype(int)
    )

    # ignore dims e.g. RGB, which we don't want to downsample
    if ignore_dims is not None:
        for dim in ignore_dims:
            slices[dim] = slice(None)

    slices = tuple(slices)

    return np.asarray(arr[slices])



from typing import *

import numpy as np
import pygfx

from ..utils import quick_min_max
from ._base import Graphic
from .features import (
    TextureArrayVolume,
    ImageCmap,
    ImageVmin,
    ImageVmax,
    ImageInterpolation,
    ImageCmapInterpolation,
    VolumeRenderMode,
    VolumeIsoThreshold,
    VolumeIsoStepSize,
    VolumeIsoSubStepSize,
    VolumeIsoEmissive,
    VolumeIsoShininess,
    VolumeSlicePlane,
    VOLUME_RENDER_MODES,
    create_volume_material_kwargs,
)


class _VolumeTile(pygfx.Volume):
    """
    Similar to pygfx.Volume, only difference is that it modifies the pick_info
    by adding the data row start indices that correspond to this chunk of the big Volume
    """

    def __init__(
        self,
        geometry,
        material,
        data_slice: tuple[slice, slice, slice],
        chunk_index: tuple[int, int, int],
        **kwargs,
    ):
        super().__init__(geometry, material, **kwargs)

        self._data_slice = data_slice
        self._chunk_index = chunk_index

    def _wgpu_get_pick_info(self, pick_value):
        pick_info = super()._wgpu_get_pick_info(pick_value)

        data_z_start, data_row_start, data_col_start = (
            self.data_slice[0].start,
            self.data_slice[1].start,
            self.data_slice[2].start,
        )

        # add the actual data row and col start indices
        x, y, z = pick_info["index"]
        x += data_col_start
        y += data_row_start
        z += data_z_start
        pick_info["index"] = (x, y, z)

        xp, yp, zp = pick_info["voxel_coord"]
        xp += data_col_start
        yp += data_row_start
        zp += data_z_start
        pick_info["voxel_coord"] = (xp, yp, zp)

        # add row chunk and col chunk index to pick_info dict
        return {
            **pick_info,
            "data_slice": self.data_slice,
            "chunk_index": self.chunk_index,
        }

    @property
    def data_slice(self) -> tuple[slice, slice, slice]:
        return self._data_slice

    @property
    def chunk_index(self) -> tuple[int, int, int]:
        return self._chunk_index


class ImageVolumeGraphic(Graphic):
    _features = {
        "data": TextureArrayVolume,
        "cmap": ImageCmap,
        "vmin": ImageVmin,
        "vmax": ImageVmax,
        "interpolation": ImageInterpolation,
        "cmap_interpolation": ImageCmapInterpolation,
        "mode": VolumeRenderMode,
        "threshold": VolumeIsoThreshold,
        "step_size": VolumeIsoStepSize,
        "substep_size": VolumeIsoSubStepSize,
        "emissive": VolumeIsoEmissive,
        "shininess": VolumeIsoShininess,
        "plane": VolumeSlicePlane,
    }

    def __init__(
        self,
        data: Any,
        mode: str = "mip",
        vmin: float = None,
        vmax: float = None,
        cmap: str = "plasma",
        interpolation: str = "linear",
        cmap_interpolation: str = "linear",
        plane: tuple[float, float, float, float] = (0, 0, -1, 0),
        threshold: float = 0.5,
        step_size: float = 1.0,
        substep_size: float = 0.1,
        emissive: str | tuple | np.ndarray = (0, 0, 0),
        shininess: int = 30,
        isolated_buffer: bool = True,
        **kwargs,
    ):
        """
        Create an ImageVolumeGraphic.

        Parameters
        ----------
        data: array-like
            array-like, usually numpy.ndarray, must support ``memoryview()``.
            Shape must be [n_planes, n_rows, n_cols] for grayscale, or [n_planes, n_rows, n_cols, 3 | 4] for RGB(A)

        mode: str, default "mip"
            render mode, one of "mip", "minip", "iso" or "slice"

        vmin: float
            lower contrast limit

        vmax: float
            upper contrast limit

        cmap: str, default "plasma"
            colormap for grayscale volumes

        interpolation: str, default "linear"
            interpolation method for sampling pixels

        cmap_interpolation: str, default "linear"
            interpolation method for sampling from colormap

        plane: (float, float, float, float), default (0, 0, -1, 0)
            Slice volume at this plane. Sets (a, b, c, d) in the equation the defines a plane: ax + by + cz + d = 0.
            Used only if `mode` = "slice"

        threshold : float, default 0.5
            The threshold texture value at which the surface is rendered.
            Used only if `mode` = "iso"

        step_size : float, default 1.0
            The size of the initial ray marching step for the initial surface finding. Smaller values will result in
            more accurate surfaces but slower rendering.
            Used only if `mode` = "iso"

        substep_size : float, default 0.1
            The size of the raymarching step for the refined surface finding. Smaller values will result in more
            accurate surfaces but slower rendering.
            Used only if `mode` = "iso"

        emissive : Color, default (0, 0, 0, 1)
            The emissive color of the surface. I.e. the color that the object emits even when not lit by a light
            source. This color is added to the final color and unaffected by lighting. The alpha channel is ignored.
            Used only if `mode` = "iso"

        shininess : int, default 30
            How shiny the specular highlight is; a higher value gives a sharper highlight.
            Used only if `mode` = "iso"

        isolated_buffer: bool, default True
            If True, initialize a buffer with the same shape as the input data and then set the data, useful if the
            data arrays are ready-only such as memmaps. If False, the input array is itself used as the
            buffer - useful if the array is large.

        kwargs
            additional keyword arguments passed to :class:`.Graphic`

        """

        valid_modes = VOLUME_RENDER_MODES.keys()
        if mode not in valid_modes:
            raise ValueError(
                f"invalid mode specified: {mode}, valid modes are: {valid_modes}"
            )

        super().__init__(**kwargs)

        world_object = pygfx.Group()

        if isinstance(data, TextureArrayVolume):
            # share existing buffer
            self._data = data
        else:
            # create new texture array to manage buffer
            # texture array that manages the textures on the GPU that represent this image volume
            self._data = TextureArrayVolume(data, isolated_buffer=isolated_buffer)

        if (vmin is None) or (vmax is None):
            _vmin, _vmax = quick_min_max(self.data.value)
            if vmin is None:
                vmin = _vmin
            if vmax is None:
                vmax = _vmax

        # other graphic features
        self._vmin = ImageVmin(vmin)
        self._vmax = ImageVmax(vmax)

        self._interpolation = ImageInterpolation(interpolation)

        # TODO: I'm assuming RGB volume images aren't supported???
        # use TextureMap for grayscale images
        self._cmap = ImageCmap(cmap)
        self._cmap_interpolation = ImageCmapInterpolation(cmap_interpolation)

        self._texture_map = pygfx.TextureMap(
            self._cmap.texture,
            filter=self._cmap_interpolation.value,
            wrap="clamp-to-edge",
        )

        self._plane = VolumeSlicePlane(plane)
        self._threshold = VolumeIsoThreshold(threshold)
        self._step_size = VolumeIsoStepSize(step_size)
        self._substep_size = VolumeIsoSubStepSize(substep_size)
        self._emissive = VolumeIsoEmissive(emissive)
        self._shininess = VolumeIsoShininess(shininess)

        material_kwargs = create_volume_material_kwargs(graphic=self, mode=mode)

        VolumeMaterialCls = VOLUME_RENDER_MODES[mode]

        self._material = VolumeMaterialCls(**material_kwargs)

        self._mode = VolumeRenderMode(mode)

        # iterate through each texture chunk and create
        # a _VolumeTile, offset the tile using the data indices
        for texture, chunk_index, data_slice in self._data:
            # create a _VolumeTile using the texture for this chunk
            vol = _VolumeTile(
                geometry=pygfx.Geometry(grid=texture),
                material=self._material,
                data_slice=data_slice,  # used to parse pick_info
                chunk_index=chunk_index,
            )

            # row and column start index for this chunk
            data_z_start = data_slice[0].start
            data_row_start = data_slice[1].start
            data_col_start = data_slice[2].start

            # offset tile position using the indices from the big data array
            # that correspond to this chunk
            vol.world.z = data_z_start
            vol.world.x = data_col_start
            vol.world.y = data_row_start

            world_object.add(vol)

        self._set_world_object(world_object)

    @property
    def data(self) -> TextureArrayVolume:
        """Get or set the image data"""
        return self._data

    @data.setter
    def data(self, data):
        self._data[:] = data

    @property
    def mode(self) -> str:
        """Get or set the volume rendering mode"""
        return self._mode.value

    @mode.setter
    def mode(self, mode: str):
        self._mode.set_value(self, mode)

    @property
    def cmap(self) -> str:
        """Get or set colormap name"""
        return self._cmap.value

    @cmap.setter
    def cmap(self, name: str):
        self._cmap.set_value(self, name)

    @property
    def vmin(self) -> float:
        """Get or set the lower contrast limit"""
        return self._vmin.value

    @vmin.setter
    def vmin(self, value: float):
        self._vmin.set_value(self, value)

    @property
    def vmax(self) -> float:
        """Get or set the upper contrast limit"""
        return self._vmax.value

    @vmax.setter
    def vmax(self, value: float):
        self._vmax.set_value(self, value)

    @property
    def interpolation(self) -> str:
        """Get or set  the image data interpolation method"""
        return self._interpolation.value

    @interpolation.setter
    def interpolation(self, value: str):
        self._interpolation.set_value(self, value)

    @property
    def cmap_interpolation(self) -> str:
        """Get or set the cmap interpolation method"""
        return self._cmap_interpolation.value

    @cmap_interpolation.setter
    def cmap_interpolation(self, value: str):
        self._cmap_interpolation.set_value(self, value)

    @property
    def plane(self) -> tuple[float, float, float, float]:
        """Get or set displayed plane in the volume. Valid only for `slice` render mode."""
        return self._plane.value

    @plane.setter
    def plane(self, value: tuple[float, float, float, float]):
        if self.mode != "slice":
            raise TypeError("`plane` property is only valid for `slice` render mode.")

        self._plane.set_value(self, value)

    @property
    def threshold(self) -> float:
        """Get or set isosurface threshold, only for `iso` mode"""
        return self._threshold.value

    @threshold.setter
    def threshold(self, value: float):
        if self.mode != "iso":
            raise TypeError(
                "`threshold` property is only used for `iso` rendering mode"
            )

        self._threshold.set_value(self, value)

    @property
    def step_size(self) -> float:
        """Get or set isosurface step_size, only for `iso` mode"""
        return self._step_size.value

    @step_size.setter
    def step_size(self, value: float):
        if self.mode != "iso":
            raise TypeError(
                "`step_size` property is only used for `iso` rendering mode"
            )

        self._step_size.set_value(self, value)

    @property
    def substep_size(self) -> float:
        """Get or set isosurface substep_size, only for `iso` mode"""
        return self._substep_size.value

    @substep_size.setter
    def substep_size(self, value: float):
        if self.mode != "iso":
            raise TypeError(
                "`substep_size` property is only used for `iso` rendering mode"
            )

        self._substep_size.set_value(self, value)

    @property
    def emissive(self) -> pygfx.Color:
        """Get or set isosurface emissive color, only for `iso` mode. Pass a <str> color, RGBA array or pygfx.Color"""
        return self._emissive.value

    @emissive.setter
    def emissive(self, value: pygfx.Color | str | tuple | np.ndarray):
        if self.mode != "iso":
            raise TypeError("`emissive` property is only used for `iso` rendering mode")

        self._emissive.set_value(self, value)

    @property
    def shininess(self) -> int:
        """Get or set isosurface shininess"""
        return self._shininess.value

    @shininess.setter
    def shininess(self, value: int):
        if self.mode != "iso":
            raise TypeError(
                "`shininess` property is only used for `iso` rendering mode"
            )

        self._shininess.set_value(self, value)

    def reset_vmin_vmax(self):
        """
        Reset the vmin, vmax by *estimating* it from the data

        Returns
        -------
        None

        """

        vmin, vmax = quick_min_max(self.data.value)
        self.vmin = vmin
        self.vmax = vmax



# This is an auto-generated file and should not be modified directly

from typing import *

import numpy

from ..graphics import *
from ..graphics._base import Graphic


class GraphicMethodsMixin:
    def _create_graphic(self, graphic_class, *args, **kwargs) -> Graphic:
        if "center" in kwargs.keys():
            center = kwargs.pop("center")
        else:
            center = False

        if "name" in kwargs.keys():
            self._check_graphic_name_exists(kwargs["name"])

        graphic = graphic_class(*args, **kwargs)
        self.add_graphic(graphic, center=center)

        return graphic

    def add_image(
        self,
        data: Any,
        vmin: float = None,
        vmax: float = None,
        cmap: str = "plasma",
        interpolation: str = "nearest",
        cmap_interpolation: str = "linear",
        isolated_buffer: bool = True,
        **kwargs,
    ) -> ImageGraphic:
        """

        Create an Image Graphic

        Parameters
        ----------
        data: array-like
            array-like, usually numpy.ndarray, must support ``memoryview()``
            | shape must be ``[n_rows, n_cols]``, ``[n_rows, n_cols, 3]`` for RGB or ``[n_rows, n_cols, 4]`` for RGBA

        vmin: float, optional
            minimum value for color scaling, estimated from data if not provided

        vmax: float, optional
            maximum value for color scaling, estimated from data if not provided

        cmap: str, optional, default "plasma"
            colormap to use to display the data. For supported colormaps see the
            ``cmap`` library catalogue: https://cmap-docs.readthedocs.io/en/stable/catalog/

        interpolation: str, optional, default "nearest"
            interpolation filter, one of "nearest" or "linear"

        cmap_interpolation: str, optional, default "linear"
            colormap interpolation method, one of "nearest" or "linear"

        isolated_buffer: bool, default True
            If True, initialize a buffer with the same shape as the input data and then
            set the data, useful if the data arrays are ready-only such as memmaps.
            If False, the input array is itself used as the buffer - useful if the
            array is large.

        kwargs:
            additional keyword arguments passed to :class:`.Graphic`


        """
        return self._create_graphic(
            ImageGraphic,
            data,
            vmin,
            vmax,
            cmap,
            interpolation,
            cmap_interpolation,
            isolated_buffer,
            **kwargs,
        )

    def add_image_volume(
        self,
        data: Any,
        mode: str = "mip",
        vmin: float = None,
        vmax: float = None,
        cmap: str = "plasma",
        interpolation: str = "linear",
        cmap_interpolation: str = "linear",
        plane: tuple[float, float, float, float] = (0, 0, -1, 0),
        threshold: float = 0.5,
        step_size: float = 1.0,
        substep_size: float = 0.1,
        emissive: str | tuple | numpy.ndarray = (0, 0, 0),
        shininess: int = 30,
        isolated_buffer: bool = True,
        **kwargs,
    ) -> ImageVolumeGraphic:
        """


        Parameters
        ----------
        data: array-like
            array-like, usually numpy.ndarray, must support ``memoryview()``.
            Shape must be [n_planes, n_rows, n_cols] for grayscale, or [n_planes, n_rows, n_cols, 3 | 4] for RGB(A)

        mode: str, default "ray"
            render mode, one of "mip", "minip", "iso" or "slice"

        vmin: float
            lower contrast limit

        vmax: float
            upper contrast limit

        cmap: str, default "plasma"
            colormap for grayscale volumes

        interpolation: str, default "linear"
            interpolation method for sampling pixels

        cmap_interpolation: str, default "linear"
            interpolation method for sampling from colormap

        plane: (float, float, float, float), default (0, 0, -1, 0)
            Slice volume at this plane. Sets (a, b, c, d) in the equation the defines a plane: ax + by + cz + d = 0.
            Used only if `mode` = "slice"

        threshold : float, default 0.5
            The threshold texture value at which the surface is rendered.
            Used only if `mode` = "iso"

        step_size : float, default 1.0
            The size of the initial ray marching step for the initial surface finding. Smaller values will result in
            more accurate surfaces but slower rendering.
            Used only if `mode` = "iso"

        substep_size : float, default 0.1
            The size of the raymarching step for the refined surface finding. Smaller values will result in more
            accurate surfaces but slower rendering.
            Used only if `mode` = "iso"

        emissive : Color, default (0, 0, 0, 1)
            The emissive color of the surface. I.e. the color that the object emits even when not lit by a light
            source. This color is added to the final color and unaffected by lighting. The alpha channel is ignored.
            Used only if `mode` = "iso"

        shininess : int, default 30
            How shiny the specular highlight is; a higher value gives a sharper highlight.
            Used only if `mode` = "iso"

        isolated_buffer: bool, default True
            If True, initialize a buffer with the same shape as the input data and then set the data, useful if the
            data arrays are ready-only such as memmaps. If False, the input array is itself used as the
            buffer - useful if the array is large.

        kwargs
            additional keyword arguments passed to :class:`.Graphic`


        """
        return self._create_graphic(
            ImageVolumeGraphic,
            data,
            mode,
            vmin,
            vmax,
            cmap,
            interpolation,
            cmap_interpolation,
            plane,
            threshold,
            step_size,
            substep_size,
            emissive,
            shininess,
            isolated_buffer,
            **kwargs,
        )

    def add_line_collection(
        self,
        data: Union[numpy.ndarray, List[numpy.ndarray]],
        thickness: Union[float, Sequence[float]] = 2.0,
        colors: Union[str, Sequence[str], numpy.ndarray, Sequence[numpy.ndarray]] = "w",
        uniform_colors: bool = False,
        cmap: Union[Sequence[str], str] = None,
        cmap_transform: Union[numpy.ndarray, List] = None,
        name: str = None,
        names: list[str] = None,
        metadata: Any = None,
        metadatas: Union[Sequence[Any], numpy.ndarray] = None,
        isolated_buffer: bool = True,
        kwargs_lines: list[dict] = None,
        **kwargs,
    ) -> LineCollection:
        """

        Create a collection of :class:`.LineGraphic`

        Parameters
        ----------
        data: list of array-like
            List or array-like of multiple line data to plot

            | if ``list`` each item in the list must be a 1D, 2D, or 3D numpy array
            | if  array-like, must be of shape [n_lines, n_points_line, y | xy | xyz]

        thickness: float or Iterable of float, default 2.0
            | if ``float``, single thickness will be used for all lines
            | if ``list`` of ``float``, each value will apply to the individual lines

        colors: str, RGBA array, Iterable of RGBA array, or Iterable of str, default "w"
            | if single ``str`` such as "w", "r", "b", etc, represents a single color for all lines
            | if single ``RGBA array`` (tuple or list of size 4), represents a single color for all lines
            | if ``list`` of ``str``, represents color for each individual line, example ["w", "b", "r",...]
            | if ``RGBA array`` of shape [data_size, 4], represents a single RGBA array for each line

        cmap: Iterable of str or str, optional
            | if ``str``, single cmap will be used for all lines
            | if ``list`` of ``str``, each cmap will apply to the individual lines

            .. note::
                ``cmap`` overrides any arguments passed to ``colors``

        cmap_transform: 1D array-like of numerical values, optional
            if provided, these values are used to map the colors from the cmap

        name: str, optional
            name of the line collection as a whole

        names: list[str], optional
            names of the individual lines in the collection, ``len(names)`` must equal ``len(data)``

        metadata: Any
            meatadata associated with the collection as a whole

        metadatas: Iterable or array
            metadata for each individual line associated with this collection, this is for the user to manage.
            ``len(metadata)`` must be same as ``len(data)``

        kwargs_lines: list[dict], optional
            list of kwargs passed to the individual lines, ``len(kwargs_lines)`` must equal ``len(data)``

        kwargs_collection
            kwargs for the collection, passed to GraphicCollection


        """
        return self._create_graphic(
            LineCollection,
            data,
            thickness,
            colors,
            uniform_colors,
            cmap,
            cmap_transform,
            name,
            names,
            metadata,
            metadatas,
            isolated_buffer,
            kwargs_lines,
            **kwargs,
        )

    def add_line(
        self,
        data: Any,
        thickness: float = 2.0,
        colors: Union[str, numpy.ndarray, Sequence] = "w",
        uniform_color: bool = False,
        cmap: str = None,
        cmap_transform: Union[numpy.ndarray, Sequence] = None,
        isolated_buffer: bool = True,
        size_space: str = "screen",
        **kwargs,
    ) -> LineGraphic:
        """

        Create a line Graphic, 2d or 3d

        Parameters
        ----------
        data: array-like
            Line data to plot. Can provide 1D, 2D, or a 3D data.
            | If passing a 1D array, it is used to set the y-values and the x-values are generated as an integer range
            from [0, data.size]
            | 2D data must be of shape [n_points, 2]. 3D data must be of shape [n_points, 3]

        thickness: float, optional, default 2.0
            thickness of the line

        colors: str, array, or iterable, default "w"
            specify colors as a single human-readable string, a single RGBA array,
            or a Sequence (array, tuple, or list) of strings or RGBA arrays

        uniform_color: bool, default ``False``
            if True, uses a uniform buffer for the line color,
            basically saves GPU VRAM when the entire line has a single color

        cmap: str, optional
            Apply a colormap to the line instead of assigning colors manually, this
            overrides any argument passed to "colors". For supported colormaps see the
            ``cmap`` library catalogue: https://cmap-docs.readthedocs.io/en/stable/catalog/

        cmap_transform: 1D array-like of numerical values, optional
            if provided, these values are used to map the colors from the cmap

        size_space: str, default "screen"
            coordinate space in which the thickness is expressed ("screen", "world", "model")

        **kwargs
            passed to :class:`.Graphic`


        """
        return self._create_graphic(
            LineGraphic,
            data,
            thickness,
            colors,
            uniform_color,
            cmap,
            cmap_transform,
            isolated_buffer,
            size_space,
            **kwargs,
        )

    def add_line_stack(
        self,
        data: List[numpy.ndarray],
        thickness: Union[float, Iterable[float]] = 2.0,
        colors: Union[str, Iterable[str], numpy.ndarray, Iterable[numpy.ndarray]] = "w",
        cmap: Union[Iterable[str], str] = None,
        cmap_transform: Union[numpy.ndarray, List] = None,
        name: str = None,
        names: list[str] = None,
        metadata: Any = None,
        metadatas: Union[Sequence[Any], numpy.ndarray] = None,
        isolated_buffer: bool = True,
        separation: float = 10.0,
        separation_axis: str = "y",
        kwargs_lines: list[dict] = None,
        **kwargs,
    ) -> LineStack:
        """

        Create a stack of :class:`.LineGraphic` that are separated along the "x" or "y" axis.

        Parameters
        ----------
        data: list of array-like
            List or array-like of multiple line data to plot

            | if ``list`` each item in the list must be a 1D, 2D, or 3D numpy array
            | if  array-like, must be of shape [n_lines, n_points_line, y | xy | xyz]

        thickness: float or Iterable of float, default 2.0
            | if ``float``, single thickness will be used for all lines
            | if ``list`` of ``float``, each value will apply to the individual lines

        colors: str, RGBA array, Iterable of RGBA array, or Iterable of str, default "w"
            | if single ``str`` such as "w", "r", "b", etc, represents a single color for all lines
            | if single ``RGBA array`` (tuple or list of size 4), represents a single color for all lines
            | if ``list`` of ``str``, represents color for each individual line, example ["w", "b", "r",...]
            | if ``RGBA array`` of shape [data_size, 4], represents a single RGBA array for each line

        cmap: Iterable of str or str, optional
            | if ``str``, single cmap will be used for all lines
            | if ``list`` of ``str``, each cmap will apply to the individual lines

            .. note::
                ``cmap`` overrides any arguments passed to ``colors``

        cmap_transform: 1D array-like of numerical values, optional
            if provided, these values are used to map the colors from the cmap

        name: str, optional
            name of the line collection as a whole

        names: list[str], optional
            names of the individual lines in the collection, ``len(names)`` must equal ``len(data)``

        metadata: Any
            metadata associated with the collection as a whole

        metadatas: Iterable or array
            metadata for each individual line associated with this collection, this is for the user to manage.
            ``len(metadata)`` must be same as ``len(data)``

        separation: float, default 10
            space in between each line graphic in the stack

        separation_axis: str, default "y"
            axis in which the line graphics in the stack should be separated


        kwargs_lines: list[dict], optional
            list of kwargs passed to the individual lines, ``len(kwargs_lines)`` must equal ``len(data)``

        kwargs_collection
            kwargs for the collection, passed to GraphicCollection


        """
        return self._create_graphic(
            LineStack,
            data,
            thickness,
            colors,
            cmap,
            cmap_transform,
            name,
            names,
            metadata,
            metadatas,
            isolated_buffer,
            separation,
            separation_axis,
            kwargs_lines,
            **kwargs,
        )

    def add_scatter(
        self,
        data: Any,
        colors: str | numpy.ndarray | tuple[float] | list[float] | list[str] = "w",
        uniform_color: bool = False,
        cmap: str = None,
        cmap_transform: numpy.ndarray = None,
        isolated_buffer: bool = True,
        sizes: Union[float, numpy.ndarray, Sequence[float]] = 1,
        uniform_size: bool = False,
        size_space: str = "screen",
        **kwargs,
    ) -> ScatterGraphic:
        """

        Create a Scatter Graphic, 2d or 3d

        Parameters
        ----------
        data: array-like
            Scatter data to plot, Can provide 2D, or a 3D data. 2D data must be of shape [n_points, 2].
            3D data must be of shape [n_points, 3]

        colors: str, array, tuple, list, Sequence, default "w"
            specify colors as a single human-readable string, a single RGBA array,
            or a Sequence (array, tuple, or list) of strings or RGBA arrays

        uniform_color: bool, default False
            if True, uses a uniform buffer for the scatter point colors. Useful if you need to
            save GPU VRAM when all points have the same color.

        cmap: str, optional
            apply a colormap to the scatter instead of assigning colors manually, this
            overrides any argument passed to "colors".  For supported colormaps see the
            ``cmap`` library catalogue: https://cmap-docs.readthedocs.io/en/stable/catalog/

        cmap_transform: 1D array-like or list of numerical values, optional
            if provided, these values are used to map the colors from the cmap

        isolated_buffer: bool, default True
            whether the buffers should be isolated from the user input array.
            Generally always ``True``, ``False`` is for rare advanced use if you have large arrays.

        sizes: float or iterable of float, optional, default 1.0
            sizes of the scatter points

        uniform_size: bool, default False
            if True, uses a uniform buffer for the scatter point sizes. Useful if you need to
            save GPU VRAM when all points have the same size.

        size_space: str, default "screen"
            coordinate space in which the size is expressed ("screen", "world", "model")

        kwargs
            passed to :class:`.Graphic`


        """
        return self._create_graphic(
            ScatterGraphic,
            data,
            colors,
            uniform_color,
            cmap,
            cmap_transform,
            isolated_buffer,
            sizes,
            uniform_size,
            size_space,
            **kwargs,
        )

    def add_text(
        self,
        text: str,
        font_size: float | int = 14,
        face_color: str | numpy.ndarray | list[float] | tuple[float] = "w",
        outline_color: str | numpy.ndarray | list[float] | tuple[float] = "w",
        outline_thickness: float = 0.0,
        screen_space: bool = True,
        offset: tuple[float] = (0, 0, 0),
        anchor: str = "middle-center",
        **kwargs,
    ) -> TextGraphic:
        """

        Create a text Graphic

        Parameters
        ----------
        text: str
            text to display

        font_size: float | int, default 10
            font size

        face_color: str, array, list, tuple, default "w"
            str or RGBA array to set the color of the text

        outline_color: str, array, list, tuple, default "w"
            str or RGBA array to set the outline color of the text

        outline_thickness: float, default 0
            relative outline thickness, value between 0.0 - 0.5

        screen_space: bool = True
            if True, text size is in screen space, if False the text size is in data space

        offset: (float, float, float), default (0, 0, 0)
            places the text at this location

        anchor: str, default "middle-center"
            position of the origin of the text
            a string representing the vertical and horizontal anchors, separated by a dash

            * Vertical values: "top", "middle", "baseline", "bottom"
            * Horizontal values: "left", "center", "right"

        **kwargs
            passed to :class:`.Graphic`


        """
        return self._create_graphic(
            TextGraphic,
            text,
            font_size,
            face_color,
            outline_color,
            outline_thickness,
            screen_space,
            offset,
            anchor,
            **kwargs,
        )



from ._base import Graphic
from .line import LineGraphic
from .scatter import ScatterGraphic
from .image import ImageGraphic
from .image_volume import ImageVolumeGraphic
from .text import TextGraphic
from .line_collection import LineCollection, LineStack


__all__ = [
    "Graphic",
    "LineGraphic",
    "ScatterGraphic",
    "ImageGraphic",
    "ImageVolumeGraphic",
    "TextGraphic",
    "LineCollection",
    "LineStack",
]



from ._linear import LinearSelector
from ._linear_region import LinearRegionSelector
from ._polygon import PolygonSelector
from ._rectangle import RectangleSelector


__all__ = ["LinearSelector", "LinearRegionSelector", "RectangleSelector"]



from ._figure import Figure
from ._subplot import Subplot
from ._utils import IMGUI

if IMGUI:
    from ._imgui_figure import ImguiFigure

    __all__ = ["Figure", "ImguiFigure"]
else:
    __all__ = ["Figure"]



from itertools import product

from math import ceil

import numpy as np

import pygfx
from ._base import GraphicFeature, GraphicFeatureEvent, block_reentrance

from ...utils import (
    make_colors,
    get_cmap_texture,
)


class TextureArray(GraphicFeature):
    """
    Manages an array of Textures representing chunks of an image.

    Creates multiple pygfx.Texture objects based on the GPU's max texture dimension limit.
    """

    event_info_spec = [
        {
            "dict key": "key",
            "type": "slice, index, numpy-like fancy index",
            "description": "key at which image data was sliced/fancy indexed",
        },
        {
            "dict key": "value",
            "type": "np.ndarray | float",
            "description": "new data values",
        },
    ]

    def __init__(self, data, isolated_buffer: bool = True):
        super().__init__()

        data = self._fix_data(data)

        shared = pygfx.renderers.wgpu.get_shared()
        self._texture_limit_2d = shared.device.limits["max-texture-dimension-2d"]

        if isolated_buffer:
            # useful if data is read-only, example: memmaps
            self._value = np.zeros(data.shape, dtype=data.dtype)
            self.value[:] = data[:]
        else:
            # user's input array is used as the buffer
            self._value = data

        # data start indices for each Texture
        self._row_indices = np.arange(
            0,
            ceil(self.value.shape[0] / self._texture_limit_2d) * self._texture_limit_2d,
            self._texture_limit_2d,
        )
        self._col_indices = np.arange(
            0,
            ceil(self.value.shape[1] / self._texture_limit_2d) * self._texture_limit_2d,
            self._texture_limit_2d,
        )

        # buffer will be an array of textures
        self._buffer: np.ndarray[pygfx.Texture] = np.empty(
            shape=(self.row_indices.size, self.col_indices.size), dtype=object
        )

        self._iter = None

        # iterate through each chunk of passed `data`
        # create a pygfx.Texture from this chunk
        for _, buffer_index, data_slice in self:
            texture = pygfx.Texture(self.value[data_slice], dim=2)

            self.buffer[buffer_index] = texture

    @property
    def value(self) -> np.ndarray:
        return self._value

    def set_value(self, graphic, value):
        self[:] = value

    @property
    def buffer(self) -> np.ndarray[pygfx.Texture]:
        return self._buffer

    @property
    def row_indices(self) -> np.ndarray:
        """
        row indices that are used to chunk the big data array
        into individual Textures on the GPU
        """
        return self._row_indices

    @property
    def col_indices(self) -> np.ndarray:
        """
        column indices that are used to chunk the big data array
        into individual Textures on the GPU
        """
        return self._col_indices

    def _fix_data(self, data):
        if data.ndim not in (2, 3):
            raise ValueError(
                "image data must be 2D with or without an RGB(A) dimension, i.e. "
                "it must be of shape [rows, cols], [rows, cols, 3] or [rows, cols, 4]"
            )

        # let's just cast to float32 always
        return data.astype(np.float32)

    def __iter__(self):
        self._iter = product(enumerate(self.row_indices), enumerate(self.col_indices))
        return self

    def __next__(self) -> tuple[pygfx.Texture, tuple[int, int], tuple[slice, slice]]:
        """
        Iterate through each Texture within the texture array

        Returns
        -------
        Texture, tuple[int, int], tuple[slice, slice]
            | Texture: pygfx.Texture
            | tuple[int, int]: chunk index, i.e corresponding index of ``self.buffer`` array
            | tuple[slice, slice]: data slice of big array in this chunk and Texture
        """
        (chunk_row, data_row_start), (chunk_col, data_col_start) = next(self._iter)

        # indices for to self.buffer for this chunk
        chunk_index = (chunk_row, chunk_col)

        # stop indices of big data array for this chunk
        row_stop = min(self.value.shape[0], data_row_start + self._texture_limit_2d)
        col_stop = min(self.value.shape[1], data_col_start + self._texture_limit_2d)

        # row and column slices that slice the data for this chunk from the big data array
        data_slice = (slice(data_row_start, row_stop), slice(data_col_start, col_stop))

        # texture for this chunk
        texture = self.buffer[chunk_index]

        return texture, chunk_index, data_slice

    def __getitem__(self, item):
        return self.value[item]

    @block_reentrance
    def __setitem__(self, key, value):
        self.value[key] = value

        for texture in self.buffer.ravel():
            texture.update_range((0, 0, 0), texture.size)

        event = GraphicFeatureEvent("data", info={"key": key, "value": value})
        self._call_event_handlers(event)

    def __len__(self):
        return self.buffer.size


class ImageVmin(GraphicFeature):
    """lower contrast limit"""

    event_info_spec = [
        {
            "dict key": "value",
            "type": "float",
            "description": "new vmin value",
        },
    ]

    def __init__(self, value: float):
        self._value = value
        super().__init__()

    @property
    def value(self) -> float:
        return self._value

    @block_reentrance
    def set_value(self, graphic, value: float):
        vmax = graphic._material.clim[1]
        graphic._material.clim = (value, vmax)
        self._value = value

        event = GraphicFeatureEvent(type="vmin", info={"value": value})
        self._call_event_handlers(event)


class ImageVmax(GraphicFeature):
    """upper contrast limit"""

    event_info_spec = [
        {
            "dict key": "value",
            "type": "float",
            "description": "new vmax value",
        },
    ]

    def __init__(self, value: float):
        self._value = value
        super().__init__()

    @property
    def value(self) -> float:
        return self._value

    @block_reentrance
    def set_value(self, graphic, value: float):
        vmin = graphic._material.clim[0]
        graphic._material.clim = (vmin, value)
        self._value = value

        event = GraphicFeatureEvent(type="vmax", info={"value": value})
        self._call_event_handlers(event)


class ImageCmap(GraphicFeature):
    """colormap for texture"""

    event_info_spec = [
        {
            "dict key": "value",
            "type": "str",
            "description": "new cmap name",
        },
    ]

    def __init__(self, value: str):
        self._value = value
        self.texture = get_cmap_texture(value)
        super().__init__()

    @property
    def value(self) -> str:
        return self._value

    @block_reentrance
    def set_value(self, graphic, value: str):
        new_colors = make_colors(256, value)
        graphic._material.map.texture.data[:] = new_colors
        graphic._material.map.texture.update_range((0, 0, 0), size=(256, 1, 1))

        self._value = value
        event = GraphicFeatureEvent(type="cmap", info={"value": value})
        self._call_event_handlers(event)


class ImageInterpolation(GraphicFeature):
    """Image interpolation method"""

    event_info_spec = [
        {
            "dict key": "value",
            "type": "str",
            "description": "new interpolation method, nearest | linear",
        },
    ]

    def __init__(self, value: str):
        self._validate(value)
        self._value = value
        super().__init__()

    def _validate(self, value):
        if value not in ["nearest", "linear"]:
            raise ValueError("`interpolation` must be one of 'nearest' or 'linear'")

    @property
    def value(self) -> str:
        return self._value

    @block_reentrance
    def set_value(self, graphic, value: str):
        self._validate(value)

        graphic._material.interpolation = value

        self._value = value
        event = GraphicFeatureEvent(type="interpolation", info={"value": value})
        self._call_event_handlers(event)


class ImageCmapInterpolation(GraphicFeature):
    """Image cmap interpolation method"""

    event_info_spec = [
        {
            "dict key": "value",
            "type": "str",
            "description": "new cmap interpolatio method, nearest | linear",
        },
    ]

    def __init__(self, value: str):
        self._validate(value)
        self._value = value
        super().__init__()

    def _validate(self, value):
        if value not in ["nearest", "linear"]:
            raise ValueError(
                "`cmap_interpolation` must be one of 'nearest' or 'linear'"
            )

    @property
    def value(self) -> str:
        return self._value

    @block_reentrance
    def set_value(self, graphic, value: str):
        self._validate(value)

        # common material for all image tiles
        graphic._material.map.min_filter = value
        graphic._material.map.mag_filter = value

        self._value = value
        event = GraphicFeatureEvent(type="cmap_interpolation", info={"value": value})
        self._call_event_handlers(event)



import logging

import numpy as np
from .mapbox_earcut import earcut as mapbox_earcut


logger = logging.getLogger("fastplotlib")


# Note: the current triangulation is in pure Python. If the results or performance of the current implementation
# proves inadequate, we can have a look at Bermuda: https://github.com/napari/bermuda


def triangulate(positions, method="earcut"):
    """Triangulate the given vertex positions.

    Returns an Nx3 integer array of faces that form a surface-mesh over the
    given positions, where N is the length of the positions minus 2,
    expressed in (local) vertex indices. The faces won't contain any
    forbidden_edges.
    """
    if len(positions) < 3:
        return np.zeros((0,), np.int32)
    if len(positions) == 3:
        return np.array([0, 1, 2], np.int32)

    # Anticipating more variations ...
    if method == "earcut":
        method = "mapbox_earcut"

    if method == "naive":
        faces = _triangulate_naive(positions)
    elif method == "mapbox_earcut":
        positions2d = positions[:, :2].flatten()
        faces = mapbox_earcut(positions2d)
        faces = np.array(faces, np.int32).reshape(-1, 3)
    else:
        raise ValueError(f"Invalid triangulation method: {method}")

    return faces


def _triangulate_naive(positions, forbidden_edges=None):
    """This tesselation algorithm simply creates edges from one vertex to all the others."""

    nverts = len(positions)
    nfaces = nverts - 2
    forbidden_edges = forbidden_edges or []

    # Determine a good point to be a reference
    forbidden_start_points = set()
    for i1, i2 in forbidden_edges:
        forbidden_start_points.add(i1)
        forbidden_start_points.add(i2)
    for i in range(len(positions)):
        if i not in forbidden_start_points:
            start_point = i
            break
    else:
        # In real meshes this cannot happen, but it can from the POV of this function's API
        raise RuntimeError("Cannot tesselate.")

    # Collect the faces
    faces = []
    i0 = start_point
    for i in range(start_point, start_point + nfaces):
        i1 = (i + 1) % nverts
        i2 = (i + 2) % nverts
        faces.append([i0, i1, i2])
    return np.array(faces, np.int32)



import numpy as np
import pygfx

from ..utils.enums import RenderQueue
from ._rect import RectManager
from ._utils import IMGUI_TOOLBAR_HEIGHT
from ..utils.types import SelectorColorStates
from ..graphics import TextGraphic


"""
Each Subplot is framed by a 2D plane mesh, a rectangle.
The rectangles are viewed using the UnderlayCamera  where (0, 0) is the top left corner.
We can control the bbox of this rectangle by changing the x and y boundaries of the rectangle.

Note how the y values of the plane mesh are negative, this is because of the UnderlayCamera.
We always just keep the positive y value, and make it negative only when setting the plane mesh.

Illustration:

(0, 0) ---------------------------------------------------
----------------------------------------------------------
----------------------------------------------------------
--------------(x0, -y0) --------------- (x1, -y0) --------
------------------------|||||||||||||||-------------------
------------------------|||||||||||||||-------------------
------------------------|||||||||||||||-------------------
------------------------|||rectangle|||-------------------
------------------------|||||||||||||||-------------------
------------------------|||||||||||||||-------------------
------------------------|||||||||||||||-------------------
--------------(x0, -y1) --------------- (x1, -y1)---------
----------------------------------------------------------
------------------------------------------- (canvas_width, canvas_height)

"""


# wgsl shader snippet for SDF function that defines the resize handler, a lower right triangle.
sdf_wgsl_resize_handle = """
// hardcode square root of 2
let m_sqrt_2 = 1.4142135;

// given a distance from an origin point, this defines the hypotenuse of a lower right triangle
let distance = (-coord.x + coord.y) / m_sqrt_2;

// return distance for this position
return distance * size;
"""


class MeshMasks:
    """Used set the x0, x1, y0, y1 positions of the plane mesh"""

    x0 = np.array(
        [
            [False, False, False],
            [True, False, False],
            [False, False, False],
            [True, False, False],
        ]
    )

    x1 = np.array(
        [
            [True, False, False],
            [False, False, False],
            [True, False, False],
            [False, False, False],
        ]
    )

    y0 = np.array(
        [
            [False, True, False],
            [False, True, False],
            [False, False, False],
            [False, False, False],
        ]
    )

    y1 = np.array(
        [
            [False, False, False],
            [False, False, False],
            [False, True, False],
            [False, True, False],
        ]
    )


masks = MeshMasks


class Frame:
    # resize handle color states
    resize_handle_color = SelectorColorStates(
        idle=(0.6, 0.6, 0.6, 1),  # gray
        highlight=(1, 1, 1, 1),  # white
        action=(1, 0, 1, 1),  # magenta
    )

    # plane color states
    plane_color = SelectorColorStates(
        idle=(0.1, 0.1, 0.1),  # dark grey
        highlight=(0.2, 0.2, 0.2),  # less dark grey
        action=(0.1, 0.1, 0.2),  # dark gray-blue
    )

    def __init__(
        self,
        viewport,
        rect,
        extent,
        resizeable,
        title,
        docks,
        toolbar_visible,
        canvas_rect,
    ):
        """
        Manages the plane mesh, resize handle point, and subplot title.
        It also sets the viewport rects for the subplot rect and the rects of the docks.

        Note: This is a backend class not meant to be user-facing.

        Parameters
        ----------
        viewport: pygfx.Viewport
            Subplot viewport

        rect: tuple | np.ndarray
            rect of this subplot

        extent: tuple | np.ndarray
            extent of this subplot

        resizeable: bool
            if the Frame is resizeable or not

        title: str
            subplot title

        docks: dict[str, PlotArea]
            subplot dock

        toolbar_visible: bool
            toolbar visibility

        canvas_rect: tuple
            figure canvas rect, the render area excluding any areas taken by imgui edge windows

        """

        self.viewport = viewport
        self.docks = docks
        self._toolbar_visible = toolbar_visible

        # create rect manager to handle all the backend rect calculations
        if rect is not None:
            self._rect_manager = RectManager(*rect, canvas_rect)
        elif extent is not None:
            self._rect_manager = RectManager.from_extent(extent, canvas_rect)
        else:
            raise ValueError("Must provide `rect` or `extent`")

        wobjects = list()

        # make title graphic
        if title is None:
            title_text = ""
        else:
            title_text = title
        self._title_graphic = TextGraphic(title_text, font_size=16, face_color="white")
        m = self._title_graphic.world_object.material
        m.alpha_mode = "blend"
        m.render_queue = RenderQueue.background
        m.depth_write = False
        m.depth_test = False
        wobjects.append(self._title_graphic.world_object)

        # init mesh of size 1 to graphically represent rect
        geometry = pygfx.plane_geometry(1, 1)
        material = pygfx.MeshBasicMaterial(
            alpha_mode="blend",
            render_queue=RenderQueue.background,
            color=self.plane_color.idle,
            depth_write=False,
            depth_test=False,
            pick_write=True,
        )
        self._plane = pygfx.Mesh(geometry, material)
        wobjects.append(self._plane)

        # Plane gets rendered before text and point
        self._plane.render_order = -1

        # create resize handler at point (x1, y1)
        x1, y1 = self.extent[[1, 3]]
        self._resize_handle = pygfx.Points(
            # note negative y since y is inverted in UnderlayCamera
            # subtract 7 so that the bottom right corner of the triangle is at the center
            pygfx.Geometry(positions=[[x1 - 7, -y1 + 7, 0]]),
            pygfx.PointsMarkerMaterial(
                alpha_mode="blend",
                render_queue=RenderQueue.background,
                color=self.resize_handle_color.idle,
                marker="custom",
                custom_sdf=sdf_wgsl_resize_handle,
                size=12,
                size_space="screen",
                depth_write=False,
                depth_test=False,
                pick_write=True,
            ),
        )

        if not resizeable:
            self._resize_handle.visible = False

        wobjects.append(self._resize_handle)

        self._world_object = pygfx.Group()
        self._world_object.add(*wobjects)

        self._reset()
        self.reset_viewport()

    @property
    def rect_manager(self) -> RectManager:
        return self._rect_manager

    @property
    def extent(self) -> np.ndarray:
        """extent, (xmin, xmax, ymin, ymax)"""
        # not actually stored, computed when needed
        return self._rect_manager.extent

    @extent.setter
    def extent(self, extent):
        self._rect_manager.extent = extent
        self._reset()
        self.reset_viewport()

    @property
    def rect(self) -> np.ndarray[int]:
        """rect in absolute screen space, (x, y, w, h)"""
        return self._rect_manager.rect

    @rect.setter
    def rect(self, rect: np.ndarray):
        self._rect_manager.rect = rect
        self._reset()
        self.reset_viewport()

    def reset_viewport(self):
        """reset the viewport rect for the subplot and docks"""

        # get rect of the render area
        x, y, w, h = self.get_render_rect()

        # dock sizes
        s_left = self.docks["left"].size
        s_top = self.docks["top"].size
        s_right = self.docks["right"].size
        s_bottom = self.docks["bottom"].size

        # top and bottom have same width
        # subtract left and right dock sizes
        w_top_bottom = w - s_left - s_right
        # top and bottom have same x pos
        x_top_bottom = x + s_left

        # set dock rects
        self.docks["left"].viewport.rect = x, y, s_left, h
        self.docks["top"].viewport.rect = x_top_bottom, y, w_top_bottom, s_top
        self.docks["bottom"].viewport.rect = (
            x_top_bottom,
            y + h - s_bottom,
            w_top_bottom,
            s_bottom,
        )
        self.docks["right"].viewport.rect = x + w - s_right, y, s_right, h

        # calc subplot rect by adjusting for dock sizes
        x += s_left
        y += s_top
        w -= s_left + s_right
        h -= s_top + s_bottom

        # set subplot rect
        self.viewport.rect = x, y, w, h

    def get_render_rect(self) -> tuple[float, float, float, float]:
        """
        Get the actual render area of the subplot, including the docks.

        Excludes area taken by the subplot title and toolbar. Also adds a small amount of spacing around the subplot.
        """
        # the rect of the entire Frame
        x, y, w, h = self.rect

        x += 1  # add 1 so a 1 pixel edge is visible
        w -= 2  # subtract 2, so we get a 1 pixel edge on both sides

        # add 4 pixels above and below title for better spacing
        y = y + 4 + self._title_graphic.font_size + 4

        # spacing on the bottom if imgui toolbar is visible
        if self.toolbar_visible:
            toolbar_space = IMGUI_TOOLBAR_HEIGHT
            resize_handle_space = 0
        else:
            toolbar_space = 0
            # need some space for resize handler if imgui toolbar isn't present
            resize_handle_space = 13

        # adjust for the 4 pixels from the line above
        # also give space for resize handler if imgui toolbar is not present
        h = (
            h
            - 4
            - self._title_graphic.font_size
            - toolbar_space
            - 4
            - resize_handle_space
        )

        return x, y, w, h

    def _reset(self):
        """reset the plane mesh using the current rect state"""

        x0, x1, y0, y1 = self._rect_manager.extent
        w = self._rect_manager.w

        self._plane.geometry.positions.data[masks.x0] = x0
        self._plane.geometry.positions.data[masks.x1] = x1

        # negative y because UnderlayCamera y is inverted
        self._plane.geometry.positions.data[masks.y0] = -y0
        self._plane.geometry.positions.data[masks.y1] = -y1

        self._plane.geometry.positions.update_full()

        # note negative y since y is inverted in UnderlayCamera
        # subtract 7 so that the bottom right corner of the triangle is at the center
        self._resize_handle.geometry.positions.data[0] = [x1 - 7, -y1 + 7, 0]
        self._resize_handle.geometry.positions.update_full()

        # set subplot title position
        x = x0 + (w / 2)
        y = y0 + (self._title_graphic.font_size / 2)
        self._title_graphic.world_object.world.x = x
        self._title_graphic.world_object.world.y = -y - 4  # add 4 pixels for spacing

    @property
    def toolbar_visible(self) -> bool:
        return self._toolbar_visible

    @toolbar_visible.setter
    def toolbar_visible(self, visible: bool):
        self._toolbar_visible = visible
        self.reset_viewport()

    @property
    def title_graphic(self) -> TextGraphic:
        return self._title_graphic

    @property
    def plane(self) -> pygfx.Mesh:
        """the plane mesh"""
        return self._plane

    @property
    def resize_handle(self) -> pygfx.Points:
        """resize handler point"""
        return self._resize_handle

    def canvas_resized(self, canvas_rect):
        """called by layout is resized"""
        self._rect_manager.canvas_resized(canvas_rect)
        self._reset()
        self.reset_viewport()



import numpy as np


"""
positions for indexing the BoxGeometry to set the "width" and "size" of the box
hacky, but I don't think we can morph meshes in pygfx yet: https://github.com/pygfx/pygfx/issues/346
"""

x_right = np.array(
    [
        True,
        True,
        True,
        True,
        False,
        False,
        False,
        False,
        False,
        True,
        False,
        True,
        True,
        False,
        True,
        False,
        False,
        True,
        False,
        True,
        True,
        False,
        True,
        False,
    ]
)

x_left = np.array(
    [
        False,
        False,
        False,
        False,
        True,
        True,
        True,
        True,
        True,
        False,
        True,
        False,
        False,
        True,
        False,
        True,
        True,
        False,
        True,
        False,
        False,
        True,
        False,
        True,
    ]
)

y_top = np.array(
    [
        False,
        True,
        False,
        True,
        False,
        True,
        False,
        True,
        True,
        True,
        True,
        True,
        False,
        False,
        False,
        False,
        False,
        False,
        True,
        True,
        False,
        False,
        True,
        True,
    ]
)

y_bottom = np.array(
    [
        True,
        False,
        True,
        False,
        True,
        False,
        True,
        False,
        False,
        False,
        False,
        False,
        True,
        True,
        True,
        True,
        True,
        True,
        False,
        False,
        True,
        True,
        False,
        False,
    ]
)

x_right = (x_right, 0)
x_left = (x_left, 0)
y_top = (y_top, 1)
y_bottom = (y_bottom, 1)



from inspect import getfullargspec
from typing import Literal, Union
from warnings import warn

import numpy as np

import pygfx
from pylinalg import vec_transform, vec_unproject
from rendercanvas import BaseRenderCanvas

from ._utils import create_controller
from ..graphics._base import Graphic
from ..graphics import ImageGraphic
from ..graphics.selectors._base_selector import BaseSelector
from ._graphic_methods_mixin import GraphicMethodsMixin
from ..legends import Legend


try:
    get_ipython()
except NameError:
    IS_IPYTHON = False
    IPYTHON = None
else:
    IS_IPYTHON = True
    IPYTHON = get_ipython()


class PlotArea(GraphicMethodsMixin):
    def __init__(
        self,
        parent: Union["PlotArea", "Figure"],
        camera: pygfx.PerspectiveCamera,
        controller: pygfx.Controller,
        scene: pygfx.Scene,
        canvas: BaseRenderCanvas,
        renderer: pygfx.WgpuRenderer,
        name: str = None,
    ):
        """
        Base class for plot creation and management. ``PlotArea`` is not intended to be instantiated by users
        but rather to provide functionality for ``subplots`` in a user ``Figure``

        Parameters
        ----------
        parent: PlotArea or Figure
            parent object

        position: Any
            position of the plot area. In a ``subplot`` position would correspond to the ``[row, column]``
            index of the ``subplot``. In docks this would correspond to a str name, "top", "right", "bottom" or "left"

        camera: pygfx.PerspectiveCamera
            Use perspective camera for both perspective and orthographic views. Set fov = 0 for orthographic projection

        controller: pygfx.Controller
            One of the pygfx controllers: "panzoom", "fly", "trackball", "orbit"

        scene: pygfx.Scene
            represents the root of a scene graph, will be viewed by the given ``camera``

        canvas: BaseRenderCanvas
            provides surface on which a scene will be rendered

        renderer: pygfx.WgpuRenderer
            renders the scene onto the canvas

        name: str, optional
            name this plot area

        """

        self._parent = parent

        self._scene = scene
        self._canvas = canvas
        self._renderer = renderer
        self._viewport: pygfx.Viewport = pygfx.Viewport(renderer)

        self._camera = camera
        self._controller = controller

        self.controller.add_camera(self._camera)
        self.controller.register_events(
            self.viewport,
        )

        self._animate_funcs_pre: list[callable] = list()
        self._animate_funcs_post: list[callable] = list()

        # list of all graphics managed by this PlotArea
        self._graphics: list[Graphic] = list()

        # selectors are in their own list so they can be excluded from scene bbox calculations
        self._selectors: list[BaseSelector] = list()

        # legends, managed just like other graphics as explained above
        self._legends: list[Legend] = list()

        # keep all graphics in a separate group, makes bbox calculations etc. easier
        # this is the "real scene" excluding axes, selection tools etc.
        self._fpl_graphics_scene = pygfx.Group()
        self.scene.add(self._fpl_graphics_scene)

        self._name = name

        # need to think about how to deal with children better
        self.children = list()

        self._background_material = pygfx.BackgroundMaterial(
            (0.0, 0.0, 0.0, 1.0),
            (0.0, 0.0, 0.0, 1.0),
            (0.0, 0.0, 0.0, 1.0),
            (0.0, 0.0, 0.0, 1.0),
            alpha_mode="blend",
        )
        self._background = pygfx.Background(None, self._background_material)
        self.scene.add(self._background)

    def get_figure(self, obj=None):
        """Get Figure instance that contains this plot area"""
        if obj is None:
            obj = self

        if obj.parent.__class__.__name__.endswith("Figure"):
            return obj.parent
        else:
            if obj.parent is None:
                raise RecursionError

            return self.get_figure(obj=obj.parent)

    # several read-only properties
    @property
    def parent(self):
        """A parent if relevant"""
        return self._parent

    @property
    def scene(self) -> pygfx.Scene:
        """The Scene where Graphics lie in this plot area"""
        return self._scene

    @property
    def canvas(self) -> BaseRenderCanvas:
        """Canvas associated to the plot area"""
        return self._canvas

    @property
    def renderer(self) -> pygfx.WgpuRenderer:
        """Renderer associated to the plot area"""
        return self._renderer

    @property
    def viewport(self) -> pygfx.Viewport:
        """The rectangular area of the renderer associated to this plot area"""
        return self._viewport

    @property
    def camera(self) -> pygfx.PerspectiveCamera:
        """camera used to view the scene"""
        return self._camera

    @camera.setter
    def camera(self, new_camera: str | pygfx.PerspectiveCamera):
        # user wants to set completely new camera, remove current camera from controller
        if isinstance(new_camera, pygfx.PerspectiveCamera):
            self.controller.remove_camera(self._camera)
            # add new camera to controller
            self.controller.add_camera(new_camera)

            self._camera = new_camera

        # modify FOV if necessary
        elif isinstance(new_camera, str):
            if new_camera == "2d":
                self._camera.fov = 0

            elif new_camera == "3d":
                # orthographic -> perspective only if fov = 0, i.e. if camera is in ortho mode
                # otherwise keep same FOV
                if self._camera.fov == 0:
                    self._camera.fov = 50

            else:
                raise ValueError(
                    "camera must be one of '2d', '3d' or a pygfx.PerspectiveCamera instance"
                )
        else:
            raise ValueError(
                "camera must be one of '2d', '3d' or a pygfx.PerspectiveCamera instance"
            )

    # in the future we can think about how to allow changing the controller
    @property
    def controller(self) -> pygfx.Controller:
        """controller used to control the camera"""
        return self._controller

    @controller.setter
    def controller(self, new_controller: str | pygfx.Controller):
        new_controller = create_controller(new_controller, self._camera)

        cameras_list = list()

        # remove all the cameras associated to this controller
        for camera in self._controller.cameras:
            self._controller.remove_camera(camera)
            cameras_list.append(camera)

        # add the associated cameras to the new controller
        for camera in cameras_list:
            new_controller.add_camera(camera)

        new_controller.register_events(self.viewport)

        # TODO: monkeypatch until we figure out a better
        #  pygfx plans on refactoring viewports anyways
        if self.parent is not None:
            if self.parent.__class__.__name__.endswith("Figure"):
                for subplot in self.parent:
                    if subplot.camera in cameras_list:
                        new_controller.register_events(subplot.viewport)
                        subplot._controller = new_controller

        self._controller = new_controller

    @property
    def graphics(self) -> tuple[Graphic, ...]:
        """Graphics in the plot area."""
        return tuple(self._graphics)

    @property
    def selectors(self) -> tuple[BaseSelector, ...]:
        """Selectors in the plot area."""
        return tuple(self._selectors)

    @property
    def legends(self) -> tuple[Legend, ...]:
        """Legends in the plot area."""
        return tuple(self._legends)

    @property
    def objects(self) -> tuple[Graphic | BaseSelector | Legend, ...]:
        return *self.graphics, *self.selectors, *self.legends

    @property
    def name(self) -> str:
        """The name of this plot area"""
        return self._name

    @name.setter
    def name(self, name: str):
        if name is None:
            self._name = None
            return

        if not isinstance(name, str):
            raise TypeError("PlotArea `name` must be of type <str>")
        self._name = name

    @property
    def background_color(self) -> tuple[pygfx.Color, ...]:
        """background colors, (top left, top right, bottom right, bottom left)"""
        return (
            self._background_material.color_top_left,
            self._background_material.color_top_right,
            self._background_material.color_bottom_right,
            self._background_material.color_bottom_left,
        )

    @background_color.setter
    def background_color(self, colors: str | tuple[float]):
        """1, 2, or 4 colors, each color must be acceptable by pygfx.Color"""
        self._background_material.set_colors(*colors)

    def map_screen_to_world(
        self, pos: tuple[float, float] | pygfx.PointerEvent, allow_outside: bool = False
    ) -> np.ndarray | None:
        """
        Map screen position to world position

        Parameters
        ----------
        pos: (float, float) | pygfx.PointerEvent
            ``(x, y)`` screen coordinates, or ``pygfx.PointerEvent``

        """
        if isinstance(pos, pygfx.PointerEvent):
            pos = pos.x, pos.y

        if not allow_outside and not self.viewport.is_inside(*pos):
            return None

        vs = self.viewport.logical_size

        # get position relative to viewport
        pos_rel = (
            pos[0] - self.viewport.rect[0],
            pos[1] - self.viewport.rect[1],
        )

        # convert screen position to NDC
        pos_ndc = (pos_rel[0] / vs[0] * 2 - 1, -(pos_rel[1] / vs[1] * 2 - 1), 0)

        # get world position
        pos_ndc += vec_transform(self.camera.world.position, self.camera.camera_matrix)
        pos_world = vec_unproject(pos_ndc[:2], self.camera.camera_matrix)

        # default z is zero for now
        return np.array([*pos_world[:2], 0])

    def _render(self):
        self._call_animate_functions(self._animate_funcs_pre)

        # does not flush, flush must be implemented in user-facing Plot objects
        self.viewport.render(self.scene, self.camera)

        for child in self.children:
            child._render()

        self._call_animate_functions(self._animate_funcs_post)

    def _call_animate_functions(self, funcs: list[callable]):
        for fn in funcs:
            try:
                args = getfullargspec(fn).args

                if len(args) > 0:
                    if args[0] == "self" and not len(args) > 1:
                        fn()
                    else:
                        fn(self)
                else:
                    fn()
            except (ValueError, TypeError):
                warn(
                    f"Could not resolve argspec of {self.__class__.__name__} animation function: {fn}, "
                    f"calling it without arguments."
                )
                fn()

    def add_animations(
        self,
        *funcs: callable,
        pre_render: bool = True,
        post_render: bool = False,
    ):
        """
        Add function(s) that are called on every render cycle.
        These are called at the Subplot level.

        Parameters
        ----------
        *funcs: callable(s)
            function(s) that are called on each render cycle

        pre_render: bool, default ``True``, optional keyword-only argument
            if true, these function(s) are called before a render cycle

        post_render: bool, default ``False``, optional keyword-only argument
            if true, these function(s) are called after a render cycle

        """
        for f in funcs:
            if not callable(f):
                raise TypeError(
                    f"all positional arguments to add_animations() must be callable types, you have passed a: {type(f)}"
                )
            if pre_render:
                self._animate_funcs_pre += funcs
            if post_render:
                self._animate_funcs_post += funcs

    def remove_animation(self, func):
        """
        Removes the passed animation function from both pre and post render.

        Parameters
        ----------
        func: callable
            The function to remove, raises a error if it's not registered as a pre or post animation function.

        """
        if func not in self._animate_funcs_pre and func not in self._animate_funcs_post:
            raise KeyError(
                f"The passed function: {func} is not registered as an animation function. These are the animation "
                f" functions that are currently registered:\n"
                f"pre: {self._animate_funcs_pre}\n\npost: {self._animate_funcs_post}"
            )

        if func in self._animate_funcs_pre:
            self._animate_funcs_pre.remove(func)

        if func in self._animate_funcs_post:
            self._animate_funcs_post.remove(func)

    def _sort_images_by_depth(self):
        """
        In general, we want to avoid setting the offset of a graphic, because the
        z-dimension may actually mean something; we cannot know whether the user is
        building a 3D scene or not. We could check whether the 3d dimension of line/point data
        is all zeros, but maybe this is intended, and *other* graphics in the same scene
        may be actually 3D. We could check camera.fov being zero, but maybe the user
        switches to a 3D camera later, or uses a 3D orthographic camera.

        The one exception, kindof, is images, which are inherently 2D, and for which
        layering helps a lot to get things rendered correctly. So we basically layer the
        images, in the order that they were added, pushing older images backwards (away
        from the camera).
        """
        count = 0
        for graphic in self._graphics:
            if isinstance(graphic, ImageGraphic):
                count += 1
                auto_depth = -count
                user_changed_depth = graphic.offset[2] % 1 > 0.0  # i.e. is not integer
                if not user_changed_depth:
                    graphic.offset = (*graphic.offset[:-1], auto_depth)

    def add_graphic(self, graphic: Graphic, center: bool = True):
        """
        Add a Graphic to the scene

        Parameters
        ----------
        graphic: Graphic or `:ref:GraphicCollection`
            Add a Graphic or a GraphicCollection to the plot area.
            Note: this must be a real Graphic instance and not a proxy

        center: bool, default True
            Center the camera on the newly added Graphic

        """

        if graphic in self:
            # graphic is already in this plot but was removed from the scene, add it back
            self._fpl_graphics_scene.add(graphic.world_object)
            return

        self._add_or_insert_graphic(graphic=graphic, center=center, action="add")

        if isinstance(graphic, ImageGraphic):
            self._sort_images_by_depth()

    def insert_graphic(
        self,
        graphic: Graphic,
        center: bool = True,
        index: int = 0,
        auto_offset: int = None,
    ):
        """
        Insert graphic into scene at given position ``index`` in stored graphics.

        Parameters
        ----------
        graphic: Graphic
            Add a Graphic to the plot area at a given position.
            Note: must be a real Graphic instance, not a weakref proxy to a Graphic

        center: bool, default True
            Center the camera on the newly added Graphic

        index: int, default 0
            Index to insert graphic.

        auto_offset: bool, default True
            If True and using an orthographic projection, sets z-axis offset of graphic to `index`

        """
        if index > len(self._graphics):
            raise IndexError(
                f"Position {index} is out of bounds for number of graphics currently "
                f"in the PlotArea: {len(self._graphics)}\n"
                f"Call `add_graphic` method to insert graphic in the last position of the stored graphics"
            )

        self._add_or_insert_graphic(
            graphic=graphic, center=center, action="insert", index=index
        )

        if isinstance(graphic, ImageGraphic):
            self._sort_images_by_depth()

    def _add_or_insert_graphic(
        self,
        graphic: Graphic,
        center: bool = True,
        action: Literal["insert", "add"] = "add",
        index: int = 0,
    ):
        """Private method to handle inserting or adding a graphic to a PlotArea."""
        if not isinstance(graphic, Graphic):
            raise TypeError(
                f"Can only add Graphic types to a PlotArea, you have passed a: {type(graphic)}"
            )

        if graphic.name is not None:  # skip for those that have no name
            self._check_graphic_name_exists(graphic.name)

        if isinstance(graphic, BaseSelector):
            obj_list = self._selectors
            self.scene.add(graphic.world_object)

        elif isinstance(graphic, Legend):
            obj_list = self._legends
            self.scene.add(graphic.world_object)

        elif isinstance(graphic, Graphic):
            obj_list = self._graphics
            self._fpl_graphics_scene.add(graphic.world_object)

            # add to tooltip registry
            if self.get_figure().show_tooltips:
                self.get_figure().tooltip_manager.register(graphic)

        else:
            raise TypeError("graphic must be of type Graphic | BaseSelector | Legend")

        if action == "insert":
            obj_list.insert(index, graphic)
        elif action == "add":
            obj_list.append(graphic)
        else:
            raise ValueError("valid actions are 'insert' | 'add'")

        if center:
            self.center_graphic(graphic)

        graphic._fpl_add_plot_area_hook(self)

    def _check_graphic_name_exists(self, name):
        if name in self:
            raise ValueError(
                f"Graphic with given name already exists in subplot or plot area. "
                f"All graphics within a subplot or plot area must have a unique name."
            )

    def center_graphic(self, graphic: Graphic, zoom: float = 1.0):
        """
        Center the camera w.r.t. the passed graphic

        Parameters
        ----------
        graphic: Graphic
            The graphic instance to center on

        zoom: float
            zoom the camera after centering

        """

        self.camera.show_object(graphic.world_object)

        # camera.show_object can cause the camera width and height to increase so apply a zoom to compensate
        # probably because camera.show_object uses bounding sphere
        self.camera.zoom = zoom

    def center_scene(self, *, zoom: float = 1.0):
        """
        Auto-center the scene, does not scale.

        Parameters
        ----------
        zoom: float
            apply a zoom after centering the scene

        """

        if not len(self._fpl_graphics_scene.children) > 0:
            return

        # scale all cameras associated with this controller
        # else it looks wonky
        for camera in self.controller.cameras:
            camera.show_object(self._fpl_graphics_scene)

            # camera.show_object can cause the camera width and height to increase so apply a zoom to compensate
            # probably because camera.show_object uses bounding sphere
            camera.zoom = zoom

    def auto_scale(
        self,
        *,  # since this is often used as an event handler, don't want to coerce maintain_aspect = True
        maintain_aspect: None | bool = None,
        zoom: float = 0.75,
    ):
        """
        Auto-scale the camera w.r.t to the scene

        Parameters
        ----------
        maintain_aspect: ``None`` or bool, default ``None``
            Maintain the camera aspect ratio for all dimensions. If ``None``, the aspect is left unchanged.
            if ``False`` the camera is scaled to the bounding box of the current scene.

        zoom: float
            zoom value for the camera after auto-scaling

        """

        if not len(self._fpl_graphics_scene.children) > 0:
            return

        self.center_scene()

        if maintain_aspect is None:  # if not provided keep current setting
            maintain_aspect = self.camera.maintain_aspect

        # scale all cameras associated with this controller else it looks wonky
        for camera in self.controller.cameras:
            camera.maintain_aspect = maintain_aspect

        if len(self._fpl_graphics_scene.children) > 0:
            width, height, depth = np.ptp(
                self._fpl_graphics_scene.get_world_bounding_box(), axis=0
            )
        else:
            width, height, depth = (1, 1, 1)

        # make sure width and height are non-zero
        if width < 0.01:
            width = 1
        if height < 0.01:
            height = 1

        # scale all cameras associated with this controller else it looks wonky
        for camera in self.controller.cameras:
            camera.width = width
            camera.height = height

            camera.zoom = zoom

    def remove_graphic(self, graphic: Graphic):
        """
        Remove a ``Graphic`` from the scene. Note: This does not garbage collect the graphic,
        you can add it back to the scene after removing it. Use ``delete_graphic()`` to
        delete and garbage collect a ``Graphic``.

        Parameters
        ----------
        graphic: Graphic
            The graphic to remove from the scene

        """

        if isinstance(graphic, (BaseSelector, Legend)):
            self.scene.remove(graphic.world_object)

        elif isinstance(graphic, Graphic):
            self._fpl_graphics_scene.remove(graphic.world_object)

    def delete_graphic(self, graphic: Graphic):
        """
        Delete the graphic, garbage collects and frees GPU VRAM.

        Parameters
        ----------
        graphic: Graphic
            The graphic to delete

        """
        if graphic not in self:
            raise KeyError(f"Graphic not found in plot area: {graphic}")

        if isinstance(graphic, BaseSelector):
            self._selectors.remove(graphic)

        elif isinstance(graphic, Legend):
            self._legends.remove(graphic)

        elif isinstance(graphic, Graphic):
            self._graphics.remove(graphic)

        # remove from scene if necessary
        if graphic.world_object in self.scene.children:
            self.scene.remove(graphic.world_object)

        elif graphic.world_object in self._fpl_graphics_scene.children:
            self._fpl_graphics_scene.remove(graphic.world_object)

        # cleanup
        graphic._fpl_prepare_del()

        if IS_IPYTHON:
            # remove any references that ipython might have made
            # check both namespaces
            for namespace in [IPYTHON.user_ns, IPYTHON.user_ns_hidden]:
                # find the reference
                for ref, obj in namespace.items():
                    if graphic is obj:
                        # we found the reference, remove from ipython
                        IPYTHON.del_var(ref)
                        break

    def clear(self):
        """
        Clear the Plot or Subplot. Also performs garbage collection, i.e. runs ``delete_graphic`` on all graphics.
        """
        for g in self.objects:
            self.delete_graphic(g)

    def __getitem__(self, name: str):
        for graphic in self.objects:
            if graphic.name == name:
                return graphic

        raise IndexError(f"No graphic or selector of given name in plot area.\n")

    def __contains__(self, item: str | Graphic):
        if isinstance(item, Graphic):
            if item in self.objects:
                return True
            else:
                return False

        elif isinstance(item, str):
            for graphic in self.objects:
                # only check named graphics
                if graphic.name is None:
                    continue

                if graphic.name == item:
                    return True

            return False

        raise TypeError("PlotArea `in` operator accepts only `Graphic` or `str` types")

    def __str__(self):
        if self.name is None:
            name = "unnamed"
        else:
            name = self.name

        return f"{name}: {self.__class__.__name__}"

    def __repr__(self):
        newline = "\n\t"

        return (
            f"{self}\n"
            f"  parent: {self.parent.__str__()}\n"
            f"  Graphics:\n"
            f"\t{newline.join(graphic.__repr__() for graphic in self.graphics)}"
            f"\n"
        )

    def __len__(self) -> int:
        return len(self._graphics) + len(self.selectors)



import warnings
from typing import *

from dataclasses import dataclass
from numbers import Real

import numpy as np
import pygfx

from ...utils.enums import RenderQueue
from .._base import Graphic
from .._collection_base import GraphicCollection
from ..features._selection_features import PolygonSelectionFeature
from ._base_selector import BaseSelector


@dataclass
class MoveInfo:
    """Movement info specific to the polygon selector."""

    # The interaction mode: None, 'create', or 'drag'
    mode: str

    # The index of the point in the polygon that is currently being manipulated
    index: int

    # The index of the point in the polygon to snap to. This is used to merge (i.e. delete) points, and to finish se the polygon.
    snap_index: int

    # The position of the cursor at the start of a drag
    start_pos: np.ndarray | None

    # The position of the vertices at the start of a drag
    start_positions: np.ndarray | None


class PolygonSelector(BaseSelector):
    _features = {"selection": PolygonSelectionFeature}

    @property
    def parent(self) -> Graphic | None:
        """Graphic that selector is associated with."""
        return self._parent

    @property
    def selection(self) -> np.ndarray[float]:
        """
        The polygon as an array of 3D points. The shape is [n_points, 3].
        """
        return self._selection.value.copy()

    @selection.setter
    def selection(self, selection: np.ndarray[float]):
        graphic = self._parent

        if isinstance(graphic, GraphicCollection):
            pass

        self._selection.set_value(self, selection)

    @property
    def limits(self) -> Tuple[float, float, float, float]:
        """Return the limits of the selector."""
        return self._limits

    @limits.setter
    def limits(self, values: Tuple[float, float, float, float]):
        if len(values) != 4 or not all(map(lambda v: isinstance(v, Real), values)):
            raise TypeError("limits must be an iterable of two numeric values")
        self._limits = tuple(
            map(round, values)
        )  # if values are close to zero things get weird so round them
        self._selection._limits = self._limits

    def __init__(
        self,
        selection: Optional[Sequence[Tuple[float]]],
        limits: Sequence[float],
        parent: Graphic = None,
        resizable: bool = True,
        fill_color=(0, 0, 0.35),
        edge_color=(0.8, 0.6, 0),
        edge_thickness: float = 4,
        vertex_color=(0.7, 0.4, 0),
        vertex_size: float = 12,
        name: str = None,
    ):
        self._parent = parent
        self._resizable = bool(resizable)

        BaseSelector.__init__(self, name=name, parent=parent)
        self._move_info = MoveInfo("none", -1, -1, None, None)

        # Initialize geometry with space for 8 points. The buffers are oversized, so we only need to create new buffers when the allocated space is full.
        # The points are 3D, even though the z-component is always 0. Indices represent the faces (i.e. the triangles).
        self.geometry = pygfx.Geometry(
            positions=np.zeros((8, 3), np.float32),
            indices=np.zeros((8, 3), np.int32),
        )

        # The draw range allows us to draw only part of the buffer, i.e. it allows us to oversize our buffers to avoid creating a new one for every added point.
        self.geometry.positions.draw_range = 0, 0
        self.geometry.indices.draw_range = 0, 0

        self._line = pygfx.Line(
            self.geometry,
            pygfx.LineMaterial(
                thickness=edge_thickness,
                color=edge_color,
                alpha_mode="blend",
                aa=True,
                render_queue=RenderQueue.selector,
                depth_test=False,
                depth_write=False,
                pick_write=True,
            ),
        )
        self._points = pygfx.Points(
            self.geometry,
            pygfx.PointsMaterial(
                size=vertex_size,
                color=vertex_color,
                alpha_mode="blend",
                aa=True,
                render_queue=RenderQueue.selector,
                depth_test=False,
                depth_write=False,
                pick_write=True,
            ),
        )
        self._indicator = pygfx.Points(
            pygfx.Geometry(positions=[[0, 0, 0]]),
            pygfx.PointsMaterial(
                size=15,
                color=vertex_color,
                alpha_mode="blend",
                opacity=0.3,
                aa=True,
                render_queue=RenderQueue.selector,
                depth_test=False,
                depth_write=False,
            ),
        )
        self._indicator.visible = False
        self._mesh = pygfx.Mesh(
            self.geometry,
            pygfx.MeshBasicMaterial(
                color=fill_color,
                alpha_mode="blend",
                opacity=0.4,
                render_queue=RenderQueue.selector,
                depth_test=False,
                depth_write=False,
                pick_write=True,
            ),
        )
        group = pygfx.Group().add(self._line, self._points, self._mesh, self._indicator)
        self._set_world_object(group)

        # Points go on top of lines, which go on top of the mesh. And indicator in between.
        self._line.render_order = 1
        self._indicator.render_order = 2
        self._points.render_order = 3

        if selection is None:
            selection = []
        self._selection = PolygonSelectionFeature(selection, (0, 0, 0, 0))

        self.edge_color = edge_color
        self.edge_width = edge_thickness
        self.limits = limits
        self.selection = self.selection  # trigger positions to be created

    def get_selected_data(
        self, graphic: Graphic = None, mode: str = "full"
    ) -> Union[np.ndarray, List[np.ndarray]]:
        """
        Get the ``Graphic`` data bounded by the current selection.
        Returns a view of the data array.

        If the ``Graphic`` is a collection, such as a ``LineStack``, it returns a list of views of the full array.
        Can be performed on the ``parent`` Graphic or on another graphic by passing to the ``graphic`` arg.

        Parameters
        ----------
        graphic: Graphic, optional, default ``None``
            if provided, returns the data selection from this graphic instead of the graphic set as ``parent``
        mode: str, default 'full'
            One of 'full', 'partial', or 'ignore'. Indicates how selected data should be returned based on the
            selectors position over the graphic. Only used for ``LineGraphic``, ``LineCollection``, and ``LineStack``
            | If 'full', will return all data bounded by the x and y limits of the selector even if partial indices
            along one axis are not fully covered by the selector.
            | If 'partial' will return only the data that is bounded by the selector, missing indices not bounded by the
            selector will be set to NaNs
            | If 'ignore', will only return data for graphics that have indices completely bounded by the selector

        Returns
        -------
        np.ndarray or List[np.ndarray]
            view or list of views of the full array, returns empty array if selection is empty
        """
        source = self._get_source(graphic)
        ixs = self.get_selected_indices(source)

        # do not need to check for mode for images, because the selector is bounded by the image shape
        # will always be `full`
        if "Image" in source.__class__.__name__:
            return source.data[ixs[:, 1], ixs[:, 0]]

        if mode not in ["full", "partial", "ignore"]:
            raise ValueError(
                f"`mode` must be one of 'full', 'partial', or 'ignore', you have passed {mode}"
            )
        if "Line" in source.__class__.__name__:
            if isinstance(source, GraphicCollection):
                data_selections: List[np.ndarray] = list()

                for i, g in enumerate(source.graphics):
                    # want to keep same length as the original line collection
                    if ixs[i].size == 0:
                        data_selections.append(
                            np.array([], dtype=np.float32).reshape(0, 3)
                        )
                    else:
                        # s gives entire slice of data along the x
                        s = slice(
                            ixs[i][0], ixs[i][-1] + 1
                        )  # add 1 because these are direct indices
                        # slices n_datapoints dim

                        # calculate missing ixs using set difference
                        # then calculate shift
                        missing_ixs = (
                            np.setdiff1d(np.arange(ixs[i][0], ixs[i][-1] + 1), ixs[i])
                            - ixs[i][0]
                        )

                        match mode:
                            # take all ixs, ignore missing
                            case "full":
                                data_selections.append(g.data[s])
                            # set missing ixs data to NaNs
                            case "partial":
                                if len(missing_ixs) > 0:
                                    data = g.data[s].copy()
                                    data[missing_ixs] = np.nan
                                    data_selections.append(data)
                                else:
                                    data_selections.append(g.data[s])
                            # ignore lines that do not have full ixs to start
                            case "ignore":
                                if len(missing_ixs) > 0:
                                    data_selections.append(
                                        np.array([], dtype=np.float32).reshape(0, 3)
                                    )
                                else:
                                    data_selections.append(g.data[s])
                return data_selections
            else:  # for lines
                if ixs.size == 0:
                    # empty selection
                    return np.array([], dtype=np.float32).reshape(0, 3)

                # add 1 to end because these are direct indices
                s = slice(ixs[0], ixs[-1] + 1)
                # slices n_datapoints dim
                # slice with min, max is faster than using all the indices

                # get missing ixs
                missing_ixs = np.setdiff1d(np.arange(ixs[0], ixs[-1] + 1), ixs) - ixs[0]

                match mode:
                    # return all, do not care about missing
                    case "full":
                        return source.data[s]
                    # set missing to NaNs
                    case "partial":
                        if len(missing_ixs) > 0:
                            data = source.data[s].copy()
                            data[missing_ixs] = np.nan
                            return data
                        else:
                            return source.data[s]
                    # missing means nothing will be returned even if selector is partially over data
                    # warn the user and return empty
                    case "ignore":
                        if len(missing_ixs) > 0:
                            warnings.warn(
                                "You have selected 'ignore' mode. Selected graphic has incomplete indices. "
                                "Move the selector or change the mode to one of `partial` or `full`."
                            )
                            return np.array([], dtype=np.float32)
                        else:
                            return source.data[s]

    def get_selected_indices(
        self, graphic: Graphic = None
    ) -> np.ndarray | tuple[np.ndarray]:
        """
        Returns the indices of the ``Graphic`` data bounded by the current selection.

        These are the data indices which correspond to the data under the selector.

        Parameters
        ----------
        graphic: Graphic, default ``None``
            If provided, returns the selection indices from this graphic instead of the graphic set as ``parent``

        Returns
        -------
        Union[np.ndarray, List[np.ndarray]]
            data indicies of the selection
            | array of (x, y) indices if the graphic is an image
            | list of indices along the x-dimension for each line if graphic is a line collection
            | array of indices along the x-dimension if graphic is a line
        """
        # get indices from source
        source = self._get_source(graphic)

        # selector (xmin, xmax, ymin, ymax) values
        polygon = self.selection[:, :2]

        # Empty ...
        if len(polygon) == 0:
            if "Image" in source.__class__.__name__:
                return np.zeros((0, 2), np.int32)
            if "Line" in source.__class__.__name__:
                if isinstance(source, GraphicCollection):
                    return [np.zeros((0, 1), np.int32) for _ in source.graphics]
                else:
                    return np.zeros((0, 1), np.int32)

        # Get bounding box to be able to do first selection
        xmin, xmax = polygon[:, 0].min(), polygon[:, 0].max()
        ymin, ymax = polygon[:, 1].min(), polygon[:, 1].max()

        # image data does not need to check for mode because the selector is always bounded
        # to the image
        if "Image" in source.__class__.__name__:
            shape = source.data.value.shape
            col_ixs = np.arange(max(0, xmin), min(xmax, shape[1] - 1), dtype=int)
            row_ixs = np.arange(max(0, ymin), min(ymax, shape[0] - 1), dtype=int)
            indices = []
            for y in row_ixs:
                for x in col_ixs:
                    p = np.array([x, y], np.float32)
                    if point_in_polygon((x, y), polygon):
                        indices.append(p)
            return np.array(indices, np.int32).reshape(-1, 2)

        if "Line" in source.__class__.__name__:
            if isinstance(source, GraphicCollection):
                ixs = list()
                for g in source.graphics:
                    points = g.data.value[:, :2] + g.offset[:2]
                    g_ixs = np.where(
                        (points[:, 0] >= xmin)
                        & (points[:, 0] <= xmax)
                        & (points[:, 1] >= ymin)
                        & (points[:, 1] <= ymax)
                    )[0]
                    g_ixs = np.array(
                        [i for i in g_ixs if point_in_polygon(points[i], polygon)],
                        g_ixs.dtype,
                    )
                    ixs.append(g_ixs)
            else:
                # map only this graphic
                points = source.data.value[:2]
                ixs = np.where(
                    (points[:, 0] >= xmin)
                    & (points[:, 0] <= xmax)
                    & (points[:, 1] >= ymin)
                    & (points[:, 1] <= ymax)
                )[0]
                ixs = np.array(
                    [i for i in ixs if point_in_polygon(points[i], polygon)],
                    ixs.dtype,
                )

            return ixs

    def _fpl_add_plot_area_hook(self, plot_area):
        self._plot_area = plot_area

        # pointer move to change endpoint of segment
        self._plot_area.renderer.add_event_handler(
            self._on_pointer_down, "pointer_down"
        )
        self._plot_area.renderer.add_event_handler(
            self._on_pointer_move, "pointer_move"
        )
        self._plot_area.renderer.add_event_handler(self._on_pointer_up, "pointer_up")

        self.position_z = len(self._plot_area) + 10

        if len(self.selection) == 0:
            self._start_move_mode("create", -1)

    def start_new_polygon(self):
        """Remove the current polygon and start drawing a new one."""
        self.selection = np.zeros((0, 3), np.float32)
        self._start_move_mode("create", -1)

    def _start_move_mode(self, what, index, start_pos=None):
        self._plot_area.controller.enabled = False
        self._move_info.mode = what
        self._move_info.index = index
        self._move_info.snap_index = None
        self._indicator.material.size = 15
        self._indicator.visible = True
        if start_pos is not None:
            self._move_info.start_pos = start_pos
            self._move_info.start_positions = self.selection.copy()
            self._indicator.visible = False

    def _end_move_mode(self):
        if self._move_info.mode == "create":
            self.world_object.children[0].material.loop = True
        self._plot_area.controller.enabled = True
        self._move_info.mode = None
        self._move_info.start_pos = None
        self._move_info.start_positions = None
        self._indicator.visible = False

    def _on_pointer_down(self, ev):
        world_pos = self._plot_area.map_screen_to_world(ev)
        if world_pos is None:
            return

        if self._move_info.mode == "create":
            # Add a polygon or finish it
            if self._move_info.snap_index is not None:
                pass  # on release we finish the polygon
            else:
                self._insert_polygon_vertex(999999, world_pos)

        elif self._move_info.mode is None:
            # Maybe initiate a drag
            if ev.target is self._points:
                index = ev.pick_info["vertex_index"]
                self._start_move_mode("drag", index)
            elif ev.target is self._line:
                index = ev.pick_info["vertex_index"]
                if ev.pick_info["segment_coord"] > 0:
                    index += 1
                self._insert_polygon_vertex(index, world_pos)
                self._start_move_mode("drag", index)
            elif ev.target is self._mesh:
                index = None  # move whole polygon
                self._start_move_mode("drag", index, world_pos)

    def _on_pointer_move(self, ev):
        """After mouse pointer move event, moves endpoint of current line segment"""
        if self._move_info.mode is None:
            return
        world_pos = self._plot_area.map_screen_to_world((ev.x, ev.y))
        if world_pos is None:
            return

        # Are we close to a point that we can snap to?
        # The concept of snapping does multiple things:
        # - preventing the user from creating points that are very close to each-other,
        # - allowing the user to finish the polygon by connecting to the start-point when in 'create' mode.
        # - allowing the user to merge points by dragging one onto its neighbour.
        index = self._move_info.index
        snap_index = None

        # Use numpy to select the nearest point.
        # This is because we cannot use picking on the actual points because
        # then we'd always pick the point being moved. We don't use a depth buffer
        # so we cannot move the point backwards to avoid it being picked.
        # An advantage is that we can make the snap-radius larger than the size of the points.
        world_pos2 = self._plot_area.map_screen_to_world((ev.x + 1, ev.y))
        world_pos_scale = float(np.linalg.norm(world_pos - world_pos2))
        snap_radius = 20  # logical screen pixels
        if len(self.selection) > 0:
            distances = np.linalg.norm(self.selection[:, :2] - world_pos[:2], axis=1)
            distances /= world_pos_scale
            distances[index] = np.inf
            snap_index = int(np.argmin(distances))
            if distances[snap_index] > snap_radius:
                snap_index = None

        if snap_index == index:  # just in case, dont snap to moving point
            snap_index = None
        if len(self.selection) < 4:
            snap_index = None
        if self._move_info.mode == "create" and snap_index != 0:
            snap_index = None
        if self._move_info.mode == "drag" and index is not None:
            last_index = len(self.selection) - 1
            if not (
                (index == 0 and snap_index == last_index)
                or (index == last_index and snap_index == 0)
                or (snap_index in (index - 1, index + 1))
            ):
                snap_index = None
        self._move_info.snap_index = snap_index

        # Show state of snap index to user
        if snap_index is not None:
            world_pos = self.geometry.positions.data[snap_index]
            self._indicator.material.size = 30
        else:
            self._indicator.material.size = 15

        self._indicator.local.position = world_pos

        # Update data
        if self._move_info.mode in ("create", "drag"):
            data = self.selection
            if len(data) > 0:
                if self._move_info.index is None:
                    delta = world_pos - self._move_info.start_pos
                    data[:] = self._move_info.start_positions + delta
                else:
                    data[self._move_info.index] = world_pos
                self._selection.set_value(self, data)

    def _on_pointer_up(self, ev):
        if self._move_info.mode in ("create", "drag"):
            # If we snapped, we dissolve (i.e. delete the vertex being moved)
            if self._move_info.snap_index is not None:
                assert self._move_info.index is not None
                self._delete_polygon_vertex(self._move_info.index)

        # Moving the mouse up may end the move action
        if self._move_info.mode == "create":
            if self._move_info.snap_index is not None:
                self._end_move_mode()
        elif self._move_info.mode == "drag":
            self._end_move_mode()

    def _insert_polygon_vertex(self, i, world_pos):
        selection = self.selection
        if len(selection) == 0:
            data = np.vstack([selection, world_pos, world_pos])
        else:
            data = np.vstack([selection[:i], world_pos, selection[i:]])
        self._selection.set_value(self, data)

    def _delete_polygon_vertex(self, i):
        selection = self.selection
        if i < 0:
            data = selection[:i]
        else:
            data = np.vstack([selection[:i], selection[i + 1 :]])
        self._selection.set_value(self, data)


def is_left(p0, p1, p2):
    """Test if point p2 is left of the line formed by p0 → p1"""
    return (p1[0] - p0[0]) * (p2[1] - p0[1]) - (p2[0] - p0[0]) * (p1[1] - p0[1])


def point_in_polygon(point, polygon):
    """Determines if the point is inside the polygon using the winding number algorithm."""
    wn = 0  # winding number counter
    n = len(polygon)

    for i in range(n):
        p0 = polygon[i]
        p1 = polygon[(i + 1) % n]

        if p0[1] <= point[1]:  # start y <= point.y
            if p1[1] > point[1]:  # upward crossing
                if is_left(p0, p1, point) > 0:
                    wn += 1  # point is left of edge
        else:  # start y > point.y
            if p1[1] <= point[1]:  # downward crossing
                if is_left(p0, p1, point) < 0:
                    wn -= 1  # point is right of edge

    return wn != 0



from dataclasses import dataclass

# this MUST be imported as early as possible in fpl.__init__ before any other wgpu stuff
from .gui import loop
from .functions import *
from .gpu import enumerate_adapters, select_adapter, print_wgpu_report
from ._plot_helpers import *
from .enums import *


@dataclass
class _Config:
    party_parrot: bool


config = _Config(party_parrot=False)



from ...layouts import IMGUI

if IMGUI:
    from ._widget import ImageWidget

else:

    class ImageWidget:
        def __init__(self, *args, **kwargs):
            raise ModuleNotFoundError(
                "ImageWidget requires `imgui-bundle` to be installed.\n"
                "pip install imgui-bundle"
            )



import enum
from typing import Literal
import numpy as np

from imgui_bundle import imgui

from ..layouts._figure import Figure


GUI_EDGES = ["right", "bottom"]


class BaseGUI:
    """
    Base class for all ImGUI based GUIs, windows and popups

    The main purpose of this base is for setting a unique ID between multiple figs with identical UI elements

    This ID can be pushed in subclasses within the `update()` method
    """

    ID_COUNTER: int = 0

    def __init__(self):
        BaseGUI.ID_COUNTER += 1
        self._id_counter = BaseGUI.ID_COUNTER

    def update(self):
        """must be implemented in subclass"""
        raise NotImplementedError


class Window(BaseGUI):
    """Base class for imgui windows drawn within Figures"""

    pass


class EdgeWindow(Window):
    def __init__(
        self,
        figure: Figure,
        size: int,
        location: Literal["bottom", "right"],
        title: str,
        window_flags: enum.IntFlag = imgui.WindowFlags_.no_collapse
        | imgui.WindowFlags_.no_resize,
        *args,
        **kwargs,
    ):
        """
        A base class for imgui windows displayed at the bottom or top edge of a Figure

        Parameters
        ----------
        figure: Figure
            Figure instance that this window will be placed in

        size: int
            width or height of the window, depending on its location

        location: str, "bottom" | "right"
            location of the window

        title: str
            window title

        window_flags: enum.IntFlag
            Window flag enum, can be compared with ``|`` operator. Valid flags are:

            .. code-block:: py

                imgui.WindowFlags_.no_title_bar
                imgui.WindowFlags_.no_resize
                imgui.WindowFlags_.no_move
                imgui.WindowFlags_.no_scrollbar
                imgui.WindowFlags_.no_scroll_with_mouse
                imgui.WindowFlags_.no_collapse
                imgui.WindowFlags_.always_auto_resize
                imgui.WindowFlags_.no_background
                imgui.WindowFlags_.no_saved_settings
                imgui.WindowFlags_.no_mouse_inputs
                imgui.WindowFlags_.menu_bar
                imgui.WindowFlags_.horizontal_scrollbar
                imgui.WindowFlags_.no_focus_on_appearing
                imgui.WindowFlags_.no_bring_to_front_on_focus
                imgui.WindowFlags_.always_vertical_scrollbar
                imgui.WindowFlags_.always_horizontal_scrollbar
                imgui.WindowFlags_.no_nav_inputs
                imgui.WindowFlags_.no_nav_focus
                imgui.WindowFlags_.unsaved_document
                imgui.WindowFlags_.no_docking
                imgui.WindowFlags_.no_nav
                imgui.WindowFlags_.no_decoration
                imgui.WindowFlags_.no_inputs

        *args
            additional args for the GUI

        **kwargs
            additional kwargs for teh GUI
        """
        super().__init__()

        if location not in GUI_EDGES:
            f"GUI does not have a valid location, valid locations are: {GUI_EDGES}, you have passed: {location}"

        self._figure = figure
        self._size = size
        self._location = location
        self._title = title
        self._window_flags = window_flags

        self._x, self._y, self._width, self._height = self.get_rect()

        self._figure.canvas.add_event_handler(self._set_rect, "resize")

    @property
    def size(self) -> int | None:
        """width or height of the edge window"""
        return self._size

    @size.setter
    def size(self, value):
        if not isinstance(value, int):
            raise TypeError
        self._size = value

    @property
    def location(self) -> str:
        """location of the window"""
        return self._location

    @property
    def x(self) -> int:
        """canvas x position of the window"""
        return self._x

    @property
    def y(self) -> int:
        """canvas y position of the window"""
        return self._y

    @property
    def width(self) -> int:
        """with the window"""
        return self._width

    @property
    def height(self) -> int:
        """height of the window"""
        return self._height

    def _set_rect(self, *args):
        self._x, self._y, self._width, self._height = self.get_rect()

    def get_rect(self) -> tuple[int, int, int, int]:
        """
        Compute the rect that defines the area this GUI is drawn to

        Returns
        -------
        int, int, int, int
            x_pos, y_pos, width, height

        """

        width_canvas, height_canvas = self._figure.canvas.get_logical_size()

        match self._location:
            case "bottom":
                x_pos = 0
                y_pos = height_canvas - self.size
                width, height = (width_canvas, self.size)

            case "right":
                x_pos, y_pos = (width_canvas - self.size, 0)
                width, height = (self.size, height_canvas)

                if self._figure.guis["bottom"] is not None:
                    height -= self._figure.guis["bottom"].size

        return x_pos, y_pos, width, height

    def draw_window(self):
        """helps simplify using imgui by managing window creation & position, and pushing/popping the ID"""
        # window position & size
        x, y, w, h = self.get_rect()
        imgui.set_next_window_size((self.width, self.height))
        imgui.set_next_window_pos((self.x, self.y))
        # imgui.set_next_window_pos((x, y))
        # imgui.set_next_window_size((w, h))
        flags = self._window_flags

        # begin window
        imgui.begin(self._title, p_open=None, flags=flags)

        # push ID to prevent conflict between multiple figs with same UI
        imgui.push_id(self._id_counter)

        # draw stuff from subclass into window
        self.update()

        # pop ID
        imgui.pop_id()

        # end the window
        imgui.end()

    def update(self):
        """Implement your GUI here and it will be drawn within the window. See the GUI examples"""
        raise NotImplementedError


class Popup(BaseGUI):
    def __init__(self, figure: Figure, *args, **kwargs):
        """
        Base class for creating ImGUI popups within Figures

        Parameters
        ----------
        figure: Figure
            Figure instance
        *args
            any args to pass to subclass constructor

        **kwargs
            any kwargs to pass to subclass constructor
        """

        super().__init__()

        self._figure = figure

        self.is_open = False

    def open(self, pos: tuple[int, int], *args, **kwargs):
        """implement in subclass"""
        raise NotImplementedError



from imgui_bundle import imgui, icons_fontawesome_6 as fa, imgui_ctx

from ..layouts._subplot import Subplot
from ._base import Window
from ..layouts._utils import IMGUI_TOOLBAR_HEIGHT


class SubplotToolbar(Window):
    def __init__(self, subplot: Subplot):
        """
        Subplot toolbar shown below all subplots
        """
        super().__init__()

        self._subplot = subplot

    def update(self):
        # get subplot rect
        x, y, width, height = self._subplot.frame.rect

        # place the toolbar window below the subplot
        pos = (x + 1, y + height - IMGUI_TOOLBAR_HEIGHT)

        imgui.set_next_window_size((width - 18, 0))
        imgui.set_next_window_pos(pos)
        flags = (
            imgui.WindowFlags_.no_collapse
            | imgui.WindowFlags_.no_title_bar
            | imgui.WindowFlags_.no_background
        )

        imgui.begin(f"Toolbar-{hex(id(self._subplot))}", p_open=None, flags=flags)

        # push ID to prevent conflict between multiple figs with same UI
        imgui.push_id(self._id_counter)
        with imgui_ctx.begin_horizontal(f"toolbar-{hex(id(self._subplot))}"):
            # autoscale button
            if imgui.button(fa.ICON_FA_MAXIMIZE):
                self._subplot.auto_scale()
            if imgui.is_item_hovered(0):
                imgui.set_tooltip("autoscale scene")

            # center scene
            if imgui.button(fa.ICON_FA_ALIGN_CENTER):
                self._subplot.center_scene()
            if imgui.is_item_hovered(0):
                imgui.set_tooltip("center scene")

            # checkbox controller
            _, self._subplot.controller.enabled = imgui.checkbox(
                fa.ICON_FA_COMPUTER_MOUSE, self._subplot.controller.enabled
            )
            if imgui.is_item_hovered(0):
                imgui.set_tooltip("enable/disable controller")

            # checkbox maintain_aspect
            _, self._subplot.camera.maintain_aspect = imgui.checkbox(
                fa.ICON_FA_EXPAND, self._subplot.camera.maintain_aspect
            )
            if imgui.is_item_hovered(0):
                imgui.set_tooltip("maintain aspect")

        # pop id when all UI has been written to window
        imgui.pop_id()

        # end window
        imgui.end()



from functools import partial

import numpy as np
import pygfx

from ._subplot import Subplot
from ._rect import RectManager


class ScreenSpaceCamera(pygfx.Camera):
    """
    Same as pygfx.ScreenCoordsCamera but y-axis is inverted.

    So top left corner is (0, 0). This is easier to manage because we
    often resize using the bottom right corner.

    """

    def _update_projection_matrix(self):
        width, height = self._view_size
        sx, sy, sz = 2 / width, 2 / height, 1
        dx, dy, dz = -1, 1, 0  # pygfx is -1, -1, 0
        m = sx, 0, 0, dx, 0, sy, 0, dy, 0, 0, sz, dz, 0, 0, 0, 1
        proj_matrix = np.array(m, dtype=float).reshape(4, 4)
        proj_matrix.flags.writeable = False
        return proj_matrix


class BaseLayout:
    def __init__(
        self,
        renderer: pygfx.WgpuRenderer,
        subplots: np.ndarray[Subplot],
        canvas_rect: tuple[float, float],
        moveable: bool,
        resizeable: bool,
    ):
        """
        Base layout engine, subclass to create a usable layout engine.
        """
        self._renderer = renderer
        self._subplots: np.ndarray[Subplot] = subplots.ravel()
        self._canvas_rect = canvas_rect

        self._last_pointer_pos: np.ndarray[np.float64, np.float64] = np.array(
            [np.nan, np.nan]
        )

        # the current user action, move or resize
        self._active_action: str | None = None
        # subplot that is currently in action, i.e. currently being moved or resized
        self._active_subplot: Subplot | None = None
        # subplot that is in focus, i.e. being hovered by the pointer
        self._subplot_focus: Subplot | None = None

        for subplot in self._subplots:
            # highlight plane when pointer enters it
            subplot.frame.plane.add_event_handler(
                partial(self._highlight_plane, subplot), "pointer_enter"
            )

            if resizeable:
                # highlight/unhighlight resize handler when pointer enters/leaves
                subplot.frame.resize_handle.add_event_handler(
                    partial(self._highlight_resize_handler, subplot), "pointer_enter"
                )
                subplot.frame.resize_handle.add_event_handler(
                    partial(self._unhighlight_resize_handler, subplot), "pointer_leave"
                )

    def _inside_render_rect(self, subplot: Subplot, pos: tuple[int, int]) -> bool:
        """whether the pos is within the render area, used for filtering out pointer events"""
        rect = subplot.frame.get_render_rect()

        x0, y0 = rect[:2]

        x1 = x0 + rect[2]
        y1 = y0 + rect[3]

        if (x0 < pos[0] < x1) and (y0 < pos[1] < y1):
            return True

        return False

    def canvas_resized(self, canvas_rect: tuple):
        """
        called by figure when canvas is resized

        Parameters
        ----------
        canvas_rect: (x, y, w, h)
            the rect that pygfx can render to, excludes any areas used by imgui.

        """

        self._canvas_rect = canvas_rect
        for subplot in self._subplots:
            subplot.frame.canvas_resized(canvas_rect)

    def _highlight_resize_handler(self, subplot: Subplot, ev):
        if self._active_action == "resize":
            return

        ev.target.material.color = subplot.frame.resize_handle_color.highlight

    def _unhighlight_resize_handler(self, subplot: Subplot, ev):
        if self._active_action == "resize":
            return

        ev.target.material.color = subplot.frame.resize_handle_color.idle

    def _highlight_plane(self, subplot: Subplot, ev):
        if self._active_action is not None:
            return

        # reset color of previous focus
        if self._subplot_focus is not None:
            self._subplot_focus.frame.plane.material.color = (
                subplot.frame.plane_color.idle
            )

        self._subplot_focus = subplot
        ev.target.material.color = subplot.frame.plane_color.highlight

    def __len__(self):
        return len(self._subplots)


class WindowLayout(BaseLayout):
    def __init__(
        self,
        renderer,
        subplots: np.ndarray[Subplot],
        canvas_rect: tuple,
        moveable=True,
        resizeable=True,
    ):
        """
        Flexible layout engine that allows freely moving and resizing subplots.
        Subplots are not allowed to overlap.

        We use a screenspace camera to perform an underlay render pass to draw the
        subplot frames, there is no depth rendering so we do not allow overlaps.

        """

        super().__init__(renderer, subplots, canvas_rect, moveable, resizeable)

        self._last_pointer_pos: np.ndarray[np.float64, np.float64] = np.array(
            [np.nan, np.nan]
        )

        for subplot in self._subplots:
            if moveable:
                # start a move action
                subplot.frame.plane.add_event_handler(
                    partial(self._action_start, subplot, "move"), "pointer_down"
                )
                # start a resize action
                subplot.frame.resize_handle.add_event_handler(
                    partial(self._action_start, subplot, "resize"), "pointer_down"
                )

        if moveable or resizeable:
            # when pointer moves, do an iteration of move or resize action
            self._renderer.add_event_handler(self._action_iter, "pointer_move")

            # end the action when pointer button goes up
            self._renderer.add_event_handler(self._action_end, "pointer_up")

    def _new_extent_from_delta(self, delta: tuple[int, int]) -> np.ndarray:
        delta_x, delta_y = delta
        if self._active_action == "resize":
            # subtract only from x1, y1
            new_extent = self._active_subplot.frame.extent - np.asarray(
                [0, delta_x, 0, delta_y]
            )
        else:
            # moving
            new_extent = self._active_subplot.frame.extent - np.asarray(
                [delta_x, delta_x, delta_y, delta_y]
            )

        x0, x1, y0, y1 = new_extent
        w = x1 - x0
        h = y1 - y0

        # make sure width and height are valid
        # min width, height is 50px
        if w <= 50:  # width > 0
            new_extent[:2] = self._active_subplot.frame.extent[:2]

        if h <= 50:  # height > 0
            new_extent[2:] = self._active_subplot.frame.extent[2:]

        # ignore movement if this would cause an overlap
        for subplot in self._subplots:
            if subplot is self._active_subplot:
                continue

            if subplot.frame.rect_manager.overlaps(new_extent):
                # we have an overlap, need to ignore one or more deltas
                # ignore x
                if not subplot.frame.rect_manager.is_left_of(
                    x0
                ) or not subplot.frame.rect_manager.is_right_of(x1):
                    new_extent[:2] = self._active_subplot.frame.extent[:2]

                # ignore y
                if not subplot.frame.rect_manager.is_above(
                    y0
                ) or not subplot.frame.rect_manager.is_below(y1):
                    new_extent[2:] = self._active_subplot.frame.extent[2:]

        # make sure all vals are non-negative
        if (new_extent[:2] < 0).any():
            # ignore delta_x
            new_extent[:2] = self._active_subplot.frame.extent[:2]

        if (new_extent[2:] < 0).any():
            # ignore delta_y
            new_extent[2:] = self._active_subplot.frame.extent[2:]

        # canvas extent
        cx0, cy0, cw, ch = self._canvas_rect

        # check if new x-range is beyond canvas x-max
        if (new_extent[:2] > cx0 + cw).any():
            new_extent[:2] = self._active_subplot.frame.extent[:2]

        # check if new y-range is beyond canvas y-max
        if (new_extent[2:] > cy0 + ch).any():
            new_extent[2:] = self._active_subplot.frame.extent[2:]

        return new_extent

    def _action_start(self, subplot: Subplot, action: str, ev):
        if self._inside_render_rect(subplot, pos=(ev.x, ev.y)):
            return

        if ev.button == 1:  # left mouse button
            self._active_action = action
            if action == "resize":
                subplot.frame.resize_handle.material.color = (
                    subplot.frame.resize_handle_color.action
                )
            elif action == "move":
                subplot.frame.plane.material.color = subplot.frame.plane_color.action
            else:
                raise ValueError

            self._active_subplot = subplot
            self._last_pointer_pos[:] = ev.x, ev.y

    def _action_iter(self, ev):
        if self._active_action is None:
            return

        delta_x, delta_y = self._last_pointer_pos - (ev.x, ev.y)
        new_extent = self._new_extent_from_delta((delta_x, delta_y))
        self._active_subplot.frame.extent = new_extent
        self._last_pointer_pos[:] = ev.x, ev.y

    def _action_end(self, ev):
        self._active_action = None
        if self._active_subplot is not None:
            self._active_subplot.frame.resize_handle.material.color = (
                self._active_subplot.frame.resize_handle_color.idle
            )
            self._active_subplot.frame.plane.material.color = (
                self._active_subplot.frame.plane_color.idle
            )
        self._active_subplot = None

        self._last_pointer_pos[:] = np.nan

    def set_rect(self, subplot: Subplot, rect: tuple | list | np.ndarray):
        """
        Set the rect of a Subplot

        Parameters
        ----------
        subplot: Subplot
            the subplot to set the rect of

        rect: (x, y, w, h)
            as absolute pixels or fractional.
            If width & height <= 1 the rect is assumed to be fractional.
            Conversely, if width & height > 1 the rect is assumed to be in absolute pixels.
            width & height must be > 0. Negative values are not allowed.

        """

        new_rect = RectManager(*rect, self._canvas_rect)
        extent = new_rect.extent
        # check for overlaps
        for s in self._subplots:
            if s is subplot:
                continue

            if s.frame.rect_manager.overlaps(extent):
                raise ValueError(f"Given rect: {rect} overlaps with another subplot.")

    def set_extent(self, subplot: Subplot, extent: tuple | list | np.ndarray):
        """
        Set the extent of a Subplot

        Parameters
        ----------
        subplot: Subplot
            the subplot to set the extent of

        extent: (xmin, xmax, ymin, ymax)
            as absolute pixels or fractional.
            If xmax & ymax <= 1 the extent is assumed to be fractional.
            Conversely, if xmax & ymax > 1 the extent is assumed to be in absolute pixels.
            Negative values are not allowed. xmax - xmin & ymax - ymin must be > 0.

        """

        new_rect = RectManager.from_extent(extent, self._canvas_rect)
        extent = new_rect.extent
        # check for overlaps
        for s in self._subplots:
            if s is subplot:
                continue

            if s.frame.rect_manager.overlaps(extent):
                raise ValueError(
                    f"Given extent: {extent} overlaps with another subplot."
                )


class GridLayout(WindowLayout):
    def __init__(
        self,
        renderer,
        subplots: np.ndarray[Subplot],
        canvas_rect: tuple[float, float, float, float],
        shape: tuple[int, int],
    ):
        """
        Grid layout engine that auto-sets Frame and Subplot rects such that they maintain
        a fixed grid layout. Does not allow freely moving or resizing subplots.

        """

        super().__init__(
            renderer, subplots, canvas_rect, moveable=False, resizeable=False
        )

        # {Subplot: (row_ix, col_ix)}, dict mapping subplots to their row and col index in the grid layout
        self._subplot_grid_position: dict[Subplot, tuple[int, int]]
        self._shape = shape

    @property
    def shape(self) -> tuple[int, int]:
        return self._shape

    def set_rect(self, subplot, rect: np.ndarray | list | tuple):
        raise NotImplementedError(
            "set_rect() not implemented for GridLayout which is an auto layout manager"
        )

    def set_extent(self, subplot, extent: np.ndarray | list | tuple):
        raise NotImplementedError(
            "set_extent() not implemented for GridLayout which is an auto layout manager"
        )

    def add_row(self):
        raise NotImplementedError("Not yet implemented")

    def add_column(self):
        raise NotImplementedError("Not yet implemented")

    def remove_row(self):
        raise NotImplementedError("Not yet implemented")

    def remove_column(self):
        raise NotImplementedError("Not yet implemented")

    def add_subplot(self):
        raise NotImplementedError(
            "Not implemented for GridLayout which is an auto layout manager"
        )

    def remove_subplot(self, subplot):
        raise NotImplementedError(
            "Not implemented for GridLayout which is an auto layout manager"
        )



from functools import partial

import numpy as np
import pygfx

from ..utils.enums import RenderQueue
from ..graphics import LineGraphic, ImageGraphic, ScatterGraphic, Graphic
from ..graphics.features import GraphicFeatureEvent


class MeshMasks:
    """Used set the x0, x1, y0, y1 positions of the plane mesh"""

    x0 = np.array(
        [
            [False, False, False],
            [True, False, False],
            [False, False, False],
            [True, False, False],
        ]
    )

    x1 = np.array(
        [
            [True, False, False],
            [False, False, False],
            [True, False, False],
            [False, False, False],
        ]
    )

    y0 = np.array(
        [
            [False, True, False],
            [False, True, False],
            [False, False, False],
            [False, False, False],
        ]
    )

    y1 = np.array(
        [
            [False, False, False],
            [False, False, False],
            [False, True, False],
            [False, True, False],
        ]
    )


masks = MeshMasks


class Tooltip:
    def __init__(self):
        # text object
        self._text = pygfx.Text(
            text="",
            font_size=12,
            screen_space=False,
            anchor="bottom-left",
            material=pygfx.TextMaterial(
                alpha_mode="blend",
                aa=True,
                render_queue=RenderQueue.overlay,
                color="w",
                outline_color="w",
                outline_thickness=0.0,
                depth_write=False,
                depth_test=False,
                pick_write=False,
            ),
        )

        # plane for the background of the text object
        geometry = pygfx.plane_geometry(1, 1)
        material = pygfx.MeshBasicMaterial(
            alpha_mode="blend",
            render_queue=RenderQueue.overlay,
            color=(0.1, 0.1, 0.3, 0.95),
            depth_write=False,
            depth_test=False,
        )
        self._plane = pygfx.Mesh(geometry, material)

        # line to outline the plane mesh
        self._line = pygfx.Line(
            geometry=pygfx.Geometry(
                positions=np.array(
                    [
                        [0, 0, 0],
                        [0, 0, 0],
                        [0, 0, 0],
                        [0, 0, 0],
                        [0, 0, 0],
                    ],
                    dtype=np.float32,
                )
            ),
            material=pygfx.LineThinMaterial(
                alpha_mode="blend",
                render_queue=RenderQueue.overlay,
                thickness=1.0,
                color=(0.8, 0.8, 1.0, 1.0),
                depth_write=False,
                depth_test=False,
            ),
        )
        # Plane gets rendered before text and line
        self._plane.render_order = -1

        self._world_object = pygfx.Group()
        self._world_object.add(self._plane, self._text, self._line)

        # padded to bbox so the background box behind the text extends a bit further
        # making the text easier to read
        self._padding = np.array([[5, 5, 0], [-5, -5, 0]], dtype=np.float32)

        self._registered_graphics = dict()

    @property
    def world_object(self) -> pygfx.Group:
        return self._world_object

    @property
    def font_size(self):
        """Get or set font size"""
        return self._text.font_size

    @font_size.setter
    def font_size(self, size: float):
        self._text.font_size = size

    @property
    def text_color(self):
        """Get or set text color using a str or RGB(A) array"""
        return self._text.material.color

    @text_color.setter
    def text_color(self, color: str | tuple | list | np.ndarray):
        self._text.material.color = color

    @property
    def background_color(self):
        """Get or set background color using a str or RGB(A) array"""
        return self._plane.material.color

    @background_color.setter
    def background_color(self, color: str | tuple | list | np.ndarray):
        self._plane.material.color = color

    @property
    def outline_color(self):
        """Get or set outline color using a str or RGB(A) array"""
        return self._line.material.color

    @outline_color.setter
    def outline_color(self, color: str | tuple | list | np.ndarray):
        self._line.material.color = color

    @property
    def padding(self) -> np.ndarray:
        """
        Get or set the background padding in number of pixels.
        The padding defines the number of pixels around the tooltip text that the background is extended by.
        """

        return self.padding[0, :2].copy()

    @padding.setter
    def padding(self, padding_xy: tuple[float, float]):
        self._padding[0, :2] = padding_xy
        self._padding[1, :2] = -np.asarray(padding_xy)

    def _set_position(self, pos: tuple[float, float]):
        """
        Set the position of the tooltip

        Parameters
        ----------
        pos: [float, float]
            position in screen space

        """
        # need to flip due to inverted y
        x, y = pos[0], pos[1]

        # put the tooltip slightly to the top right of the cursor positoin
        x += 8
        y -= 8

        self._text.world.position = (x, -y, 0)

        bbox = self._text.get_world_bounding_box() - self._padding
        [[x0, y0, _], [x1, y1, _]] = bbox

        self._plane.geometry.positions.data[masks.x0] = x0
        self._plane.geometry.positions.data[masks.x1] = x1
        self._plane.geometry.positions.data[masks.y0] = y0
        self._plane.geometry.positions.data[masks.y1] = y1

        self._plane.geometry.positions.update_range()

        # line points
        pts = [[x0, y0], [x0, y1], [x1, y1], [x1, y0], [x0, y0]]

        self._line.geometry.positions.data[:, :2] = pts
        self._line.geometry.positions.update_range()

    def _event_handler(self, custom_tooltip: callable, ev: pygfx.PointerEvent):
        """Handles the tooltip appear event, determines the text to be set in the tooltip"""
        if custom_tooltip is not None:
            info = custom_tooltip(ev)

        elif isinstance(ev.graphic, ImageGraphic):
            col, row = ev.pick_info["index"]
            if ev.graphic.data.value.ndim == 2:
                info = str(ev.graphic.data[row, col])
            else:
                info = "\n".join(
                    f"{channel}: {val}"
                    for channel, val in zip("rgba", ev.graphic.data[row, col])
                )

        elif isinstance(ev.graphic, (LineGraphic, ScatterGraphic)):
            index = ev.pick_info["vertex_index"]
            info = "\n".join(
                f"{dim}: {val}" for dim, val in zip("xyz", ev.graphic.data[index])
            )
        else:
            raise TypeError("Unsupported graphic")

        # make the tooltip object visible
        self.world_object.visible = True

        # set the text and top left position of the tooltip
        self._text.set_text(info)
        self._set_position((ev.x, ev.y))

    def _clear(self, ev):
        self._text.set_text("")
        self.world_object.visible = False

    def register(
        self,
        graphic: Graphic,
        appear_event: str = "pointer_move",
        disappear_event: str = "pointer_leave",
        custom_info: callable = None,
    ):
        """
        Register a Graphic to display tooltips.

        **Note:** if the passed graphic is already registered then it first unregistered
        and then re-registered using the given arguments.

        Parameters
        ----------
        graphic: Graphic
            Graphic to register

        appear_event: str, default "pointer_move"
            the pointer that triggers the tooltip to appear. Usually one of "pointer_move" | "click" | "double_click"

        disappear_event: str, default "pointer_leave"
            the event that triggers the tooltip to disappear, does not have to be a pointer event.

        custom_info: callable, default None
            a custom function that takes the pointer event defined as the `appear_event` and returns the text
            to display in the tooltip

        """
        if graphic in list(self._registered_graphics.keys()):
            # unregister first and then re-register
            self.unregister(graphic)

        pfunc = partial(self._event_handler, custom_info)
        graphic.add_event_handler(pfunc, appear_event)
        graphic.add_event_handler(self._clear, disappear_event)

        self._registered_graphics[graphic] = (pfunc, appear_event, disappear_event)

        # automatically unregister when graphic is deleted
        graphic.add_event_handler(self.unregister, "deleted")

    def unregister(self, graphic: Graphic):
        """
        Unregister a Graphic to no longer display tooltips for this graphic.

        **Note:** if the passed graphic is not registered then it is just ignored without raising any exception.

        Parameters
        ----------
        graphic: Graphic
            Graphic to unregister

        """

        if isinstance(graphic, GraphicFeatureEvent):
            # this happens when the deleted event is triggered
            graphic = graphic.graphic

        if graphic not in self._registered_graphics:
            return

        # get pfunc and event names
        pfunc, appear_event, disappear_event = self._registered_graphics.pop(graphic)

        # remove handlers from graphic
        graphic.remove_event_handler(pfunc, appear_event)
        graphic.remove_event_handler(self._clear, disappear_event)

    def unregister_all(self):
        """unregister all graphics"""
        for graphic in self._registered_graphics.keys():
            self.unregister(graphic)
